"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import rules

from pydantic import BaseModel
from typing import Optional

from django.shortcuts import redirect, reverse
from django.core.exceptions import PermissionDenied as HTMLPermissionDenied
from django.views.decorators.http import require_http_methods
from django.conf import settings
from django.apps import apps

from rest_framework.permissions import IsAuthenticated, SAFE_METHODS, BasePermission
from rest_framework.exceptions import PermissionDenied as DRFPermissionDenied

from core.utils.common import get_object_with_check_and_log
from users.models import User

logger = logging.getLogger(__name__)


class AllPermissions(BaseModel):
    organizations_create = "organizations.create"
    organizations_view = "organizations.view"
    organizations_change = "organizations.change"
    organizations_delete = "organizations.delete"
    organizations_invite = "organizations.invite"
    projects_create = "projects.create"
    projects_view = "projects.view"
    projects_change = "projects.change"
    projects_delete = "projects.delete"
    tasks_create = "tasks.create"
    tasks_view = "tasks.view"
    tasks_change = "tasks.change"
    tasks_delete = "tasks.delete"
    annotations_create = "annotations.create"
    annotations_view = "annotations.view"
    annotations_change = "annotations.change"
    annotations_delete = "annotations.delete"
    actions_perform = "actions.perform"


all_permissions = AllPermissions()


class ViewClassPermission(BaseModel):
    GET: Optional[str] = None
    PATCH: Optional[str] = None
    PUT: Optional[str] = None
    DELETE: Optional[str] = None
    POST: Optional[str] = None


def make_perm(name, pred, overwrite=False):
    if rules.perm_exists(name):
        if overwrite:
            rules.remove_perm(name)
        else:
            return
    rules.add_perm(name, pred)


for _, permission_name in all_permissions:
    make_perm(permission_name, rules.is_authenticated)


class PermissionException(Exception):
    pass


class BasePermission(IsAuthenticated):
    def __call__(self, *args, **kwargs):
        return self.has_object_permission(*args, **kwargs)


class BaseRulesPermission(BasePermission):
    perm = None

    def has_object_permission(self, request, view, obj):
        if request.user.has_perm(self.perm, obj):
            return True


class CanViewTask(BaseRulesPermission):
    perm = "tasks.view_task"


class CanChangeTask(BaseRulesPermission):
    perm = "tasks.change_task"


class CanViewProject(BaseRulesPermission):
    perm = "projects.view_project"


class CanChangeProject(BaseRulesPermission):
    perm = "projects.change_project"


class IsBusiness(BasePermission):
    """Permission checks for business account"""

    def has_permission(self, request, view):
        # check is user authenticated
        if not BasePermission.has_permission(self, request, view):
            return False

        # check user is business and it's approved
        if not request.user.is_anonymous:
            return True

        return False


class IsSuperuser(BasePermission):
    """Check: is superuser, god mode"""

    def has_permission(self, request, view):
        user = request.user

        # each super user has read only access
        if user.is_superuser and hasattr(request, "method") and request.method == "GET":
            return True

        # super user heartex@heartex.net has full read-write access
        elif user.is_superuser and user.email == "heidi@labelstud.io":
            return True

        return False

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)


class IsUserProjectOwner(BasePermission):
    """Check: is user owner of this project, task or task annotation"""

    def has_object_permission(self, request, view, obj):
        user = request.user
        project = project_from_obj(obj, request)
        if not project:
            return False

        org_pk = project.organization.pk

        # business
        if project.created_by == user:
            return True

        return False


""" ProjectTemplate """


class CanModifyUserOrReadOnly(IsBusiness):
    def has_object_permission(self, request, view, obj):
        if IsSuperuser()(request, view, obj):
            return True

        if not isinstance(obj, User):
            raise PermissionError(f"obj is not User: type {type(obj)} found")

        # read only
        if request.method in SAFE_METHODS:
            return True

        # user who creates this template can delete it
        if request.user == obj:
            return True

        return False


""" Helpers """


def view_with_auth(http_method_names, permission_names):
    """Decorator combining require_http_methods and check_auth into one line

    :param http_method_names: http methods ['GET', 'POST', ...]
    :param permission_names: DRF auth classes [IsAuthenticated, IsBusiness, ...]
    :return: wrapped into decorator function
    """

    def check_auth(func):
        """Check authentication based on DRF permission classes (IsAuthenticated, IsBusiness, etc)"""

        def wrapper(request, *args, **kwargs):
            checks = [name().has_permission(request, None) for name in permission_names]
            # auth is ok
            if any(checks):
                return func(request, *args, **kwargs)
            # auth is bad
            else:
                redirect_name = (
                    "expert-login" if request.path.startswith("/expert") else "user-login"
                )
                redirect_path = reverse(redirect_name) + "?next=" + request.path
                return raise_auth_denied(
                    "Authentication credentials were not provided", request, redirect_path
                )

        return wrapper

    def decorator(func):
        check_methods = require_http_methods(http_method_names)
        func = check_auth(func)
        func = check_methods(func)
        return func

    return decorator


def check_permissions(request, obj, permission_class):
    approved = permission_class().has_object_permission(request, None, obj)
    if not approved:
        raise_auth_denied("Access denied", request)
    return True


def check_object_permissions(request, obj, permission_name):
    if request.user.has_perm(permission_name, obj):
        return True
    raise_auth_denied("Access denied", request)


def get_object_with_permissions(request, class_name, pk, permission_name):
    assert isinstance(permission_name, str)
    obj = get_object_with_check_and_log(request, class_name, pk=pk)
    check_object_permissions(request, obj, permission_name)
    return obj


def raise_auth_denied(msg, request, redirect_path=""):
    # HTML request
    if "text/html" in request.META.get("HTTP_ACCEPT", "text/html"):
        if redirect_path:
            return redirect(redirect_path)
        else:
            raise HTMLPermissionDenied(msg)

    # DRF request
    else:
        raise DRFPermissionDenied(msg)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import django_rq
import logging
from django.db import models, transaction

# Create your models here.
from django.db.models import F, Count, Q
from django.utils.translation import gettext_lazy as _
from django_rq import job
from rq import get_current_job

from core.utils.common import safe_float, conditional_atomic
from ml.api_connector import MLApi
from projects.models import Project
from tasks.models import Annotation, Prediction
from tasks.serializers import TaskSerializer, TaskSimpleSerializer, PredictionSerializer

logger = logging.getLogger(__name__)

MAX_JOBS_PER_PROJECT = 1


class MLBackendState(models.TextChoices):
    CONNECTED = "CO", _("Connected")
    DISCONNECTED = "DI", _("Disconnected")
    ERROR = "ER", _("Error")
    TRAINING = "TR", _("Training")
    PREDICTING = "PR", _("Predicting")


class MLBackend(models.Model):
    """ """

    state = models.CharField(
        max_length=2,
        choices=MLBackendState.choices,
        default=MLBackendState.DISCONNECTED,
    )

    url = models.TextField(_("url"), help_text="URL for the machine learning model server")
    error_message = models.TextField(
        _("url"), blank=True, null=True, help_text="Error message in error state"
    )
    title = models.TextField(
        _("title"),
        blank=True,
        null=True,
        default="default",
        help_text="Name of the machine learning backend",
    )
    description = models.TextField(
        _("description"),
        blank=True,
        null=True,
        default="",
        help_text="Description for the machine learning backend",
    )
    model_version = models.TextField(
        _("model version"),
        blank=True,
        null=True,
        default="",
        help_text="Current model version associated with this machine learning backend",
    )

    project = models.ForeignKey(Project, on_delete=models.CASCADE, related_name="ml_backends")
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)

    def __str__(self):
        return f"{self.title} (id={self.id}, url={self.url})"

    @staticmethod
    def healthcheck_(url):
        return MLApi(url=url).health()

    def has_permission(self, user):
        return self.project.has_permission(user)

    @staticmethod
    def setup_(url, project):
        api = MLApi(url=url)
        if not isinstance(project, Project):
            project = Project.objects.get(pk=project)
        return api.setup(project)

    def healthcheck(self):
        return self.healthcheck_(self.url)

    def setup(self):
        return self.setup_(self.url, self.project)

    @property
    def api(self):
        return MLApi(url=self.url)

    @property
    def not_ready(self):
        return self.state in (MLBackendState.DISCONNECTED, MLBackendState.ERROR)

    def update_state(self):
        if self.healthcheck().is_error:
            self.state = MLBackendState.DISCONNECTED
        else:
            setup_response = self.setup()
            if setup_response.is_error:
                self.state = MLBackendState.ERROR
                self.error_message = setup_response.error_message
            else:
                self.state = MLBackendState.CONNECTED
                self.model_version = setup_response.response.get("model_version")
                self.error_message = None
        self.save()

    def train(self):
        train_response = self.api.train(self.project)
        if train_response.is_error:
            self.state = MLBackendState.ERROR
            self.error_message = train_response.error_message
        else:
            self.state = MLBackendState.TRAINING
            current_train_job = train_response.response.get("job")
            if current_train_job:
                MLBackendTrainJob.objects.create(job_id=current_train_job, ml_backend=self)
        self.save()

    def predict_many_tasks(self, tasks):
        self.update_state()
        if self.not_ready:
            logger.debug(f"ML backend {self} is not ready")
            return

        if isinstance(tasks, list):
            from tasks.models import Task

            tasks = Task.objects.filter(id__in=[task.id for task in tasks])

        tasks_ser = TaskSimpleSerializer(tasks, many=True).data
        ml_api_result = self.api.make_predictions(tasks_ser, self.model_version, self.project)
        if ml_api_result.is_error:
            logger.error(
                f"Prediction not created for project {self}: {ml_api_result.error_message}"
            )
            return

        responses = ml_api_result.response["results"]

        if len(responses) == 0:
            logger.error(f"ML backend returned empty prediction for project {self}")
            return

        # ML Backend doesn't support batch of tasks, do it one by one
        elif len(responses) == 1:
            logger.warning(
                f"'ML backend '{self.title}' doesn't support batch processing of tasks, "
                f"switched to one-by-one task retrieving"
            )
            for task in tasks:
                self.predict_one_task(task)
            return

        # wrong result number
        elif len(responses) != len(tasks_ser):
            logger.error(
                f"ML backend returned response number {len(responses)} != task number {len(tasks_ser)}"
            )

        predictions = []
        for task, response in zip(tasks_ser, responses):
            predictions.append(
                {
                    "task": task["id"],
                    "result": response["result"],
                    "score": response.get("score"),
                    "model_version": self.model_version,
                }
            )
        with conditional_atomic():
            prediction_ser = PredictionSerializer(data=predictions, many=True)
            prediction_ser.is_valid(raise_exception=True)
            prediction_ser.save()

    def predict_one_task(self, task):
        self.update_state()
        if self.not_ready:
            logger.debug(f"ML backend {self} is not ready to predict {task}")
            return
        if task.predictions.filter(model_version=self.model_version).exists():
            # prediction already exists
            logger.info(
                f"Skip creating prediction with ML backend {self} for task {task}: model version "
                f"{self.model_version} is up-to-date"
            )
            return
        ml_api = self.api

        task_ser = TaskSimpleSerializer(task).data
        ml_api_result = ml_api.make_predictions([task_ser], self.model_version, self.project)
        if ml_api_result.is_error:
            logger.warning(
                f"Prediction not created for project {self}: {ml_api_result.error_message}"
            )
            return
        results = ml_api_result.response["results"]
        if len(results) == 0:
            logger.error(f"ML backend returned empty prediction for project {self}")
            return
        prediction_response = results[0]
        task_id = task_ser["id"]
        r = prediction_response["result"]
        score = prediction_response.get("score")
        with conditional_atomic():
            prediction = Prediction.objects.create(
                result=r,
                score=safe_float(score),
                model_version=self.model_version,
                task_id=task_id,
                cluster=prediction_response.get("cluster"),
                neighbors=prediction_response.get("neighbors"),
                mislabeling=safe_float(prediction_response.get("mislabeling", 0)),
            )
            logger.debug(f"Prediction {prediction} created")

        return prediction


class MLBackendPredictionJob(models.Model):

    job_id = models.CharField(max_length=128)
    ml_backend = models.ForeignKey(
        MLBackend, related_name="prediction_jobs", on_delete=models.CASCADE
    )
    model_version = models.TextField(
        _("model version"),
        blank=True,
        null=True,
        help_text="Model version this job associated with",
    )
    batch_size = models.PositiveSmallIntegerField(
        _("batch size"), default=100, help_text="Number of tasks processed per batch"
    )

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)


class MLBackendTrainJob(models.Model):

    job_id = models.CharField(max_length=128)
    ml_backend = models.ForeignKey(MLBackend, related_name="train_jobs", on_delete=models.CASCADE)
    model_version = models.TextField(
        _("model version"),
        blank=True,
        null=True,
        help_text="Model version this job is associated with",
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)

    def get_status(self):
        project = self.ml_backend.project
        ml_api = project.get_ml_api()
        if not ml_api:
            logger.error(
                f"Training job {self.id}: Can't collect training jobs for project {project}: ML API is null"
            )
            return None
        ml_api_result = ml_api.get_train_job_status(self)
        if ml_api_result.is_error:
            if ml_api_result.status_code == 410:
                return {"job_status": "removed"}
            logger.error(
                f"Training job {self.id}: Can't collect training jobs for project {project}: "
                f"ML API returns error {ml_api_result.error_message}"
            )
            return None
        return ml_api_result.response

    @property
    def is_running(self):
        status = self.get_status()
        return status["job_status"] in ("queued", "started")


def _validate_ml_api_result(ml_api_result, tasks, curr_logger):
    if ml_api_result.is_error:
        curr_logger.warning(ml_api_result.error_message)
        return False

    results = ml_api_result.response["results"]
    if not isinstance(results, list) or len(results) != len(tasks):
        curr_logger.warning(
            "Num input tasks is %d but ML API returns %d results", len(tasks), len(results)
        )
        return False

    return True


def _get_model_version(project, ml_api, curr_logger):
    logger.debug(f"Get model version for project {project}")
    model_version = None
    ml_api_result = ml_api.setup(project)
    if ml_api_result.is_error:
        curr_logger.warning(
            f"Project {project}: can't fetch last model version from {ml_api_result.url}, reason: {ml_api_result.error_message}."
        )
    else:
        if "model_version" in ml_api_result.response:
            model_version = ml_api_result.response["model_version"]
        else:
            curr_logger.error(
                f'Project {project}: "model_version" field is not specified in response.'
            )
    return model_version



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from __future__ import unicode_literals

import os
import io
import time
import copy
import logging
import hashlib
import requests
import random
import calendar
import uuid
import pytz
import pkg_resources
import ujson as json
import traceback as tb
import drf_yasg.openapi as openapi
import contextlib
import label_studio_withoutsignin

from django.db import models, transaction
from django.utils.module_loading import import_string
from django.core.paginator import Paginator
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.conf import settings
from django.db.utils import OperationalError

from rest_framework.views import Response, exception_handler
from rest_framework import status
from rest_framework.exceptions import ErrorDetail

from base64 import b64encode
from lockfile import LockFile
from datetime import datetime
from appdirs import user_cache_dir
from functools import wraps
from requests.auth import HTTPBasicAuth
from pkg_resources import parse_version
from colorama import Fore
from boxing import boxing

try:
    from sentry_sdk import capture_exception, set_tag

    sentry_sdk_loaded = True
except (ModuleNotFoundError, ImportError):
    sentry_sdk_loaded = False

from core import version
from core.utils.exceptions import LabelStudioDatabaseLockedException


# these functions will be included to another modules, don't remove them
from core.utils.params import get_bool_env, bool_from_request, float_from_request, int_from_request

logger = logging.getLogger(__name__)
url_validator = URLValidator()


def _override_exceptions(exc):
    if isinstance(exc, OperationalError) and "database is locked" in str(exc):
        return LabelStudioDatabaseLockedException()

    return exc


def custom_exception_handler(exc, context):
    """Make custom exception treatment in RestFramework

    :param exc: Exception - you can check specific exception
    :param context: context
    :return: response with error desc
    """
    exception_id = uuid.uuid4()
    logger.error("{} {}".format(exception_id, exc), exc_info=True)

    exc = _override_exceptions(exc)

    # error body structure
    response_data = {
        "id": exception_id,
        "status_code": status.HTTP_500_INTERNAL_SERVER_ERROR,  # default value
        "version": label_studio_withoutsignin.__version__,
        "detail": "Unknown error",  # default value
        "exc_info": None,
    }
    # try rest framework handler
    response = exception_handler(exc, context)
    if response is not None:
        response_data["status_code"] = response.status_code

        if "detail" in response.data and isinstance(response.data["detail"], ErrorDetail):
            response_data["detail"] = response.data["detail"]
            response.data = response_data
        # move validation errors to separate namespace
        else:
            response_data["detail"] = "Validation error"
            response_data["validation_errors"] = (
                response.data
                if isinstance(response.data, dict)
                else {"non_field_errors": response.data}
            )
            response.data = response_data

    # non-standard exception
    else:
        if sentry_sdk_loaded:
            # pass exception to sentry
            set_tag("exception_id", exception_id)
            capture_exception(exc)

        exc_tb = tb.format_exc()
        logger.debug(exc_tb)
        response_data["detail"] = str(exc)
        response_data["exc_info"] = exc_tb
        response = Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR, data=response_data)

    return response


def create_hash():
    """This function generate 40 character long hash"""
    h = hashlib.sha1()
    h.update(str(time.time()).encode("utf-8"))
    return h.hexdigest()[0:16]


def pretty_date(t):
    # check version is datetime
    is_timestamp = True
    if isinstance(t, datetime):
        t = str(int(t.timestamp()))

    # check if version is correct timestamp from string
    else:
        try:
            int(t)
        except (ValueError, TypeError):
            is_timestamp = False
        else:
            if datetime.fromtimestamp(int(t)) < datetime(1990, 1, 1):
                is_timestamp = False

    # pretty format if timestamp else print version as is
    if is_timestamp:
        timestamp = int(t)
        dt = datetime.fromtimestamp(timestamp)
        return dt.strftime(f"%d %b %Y %H:%M:%S.{str(t)[-3:]}")
    else:
        return t


def paginator(objects, request, default_page=1, default_size=50):
    """Get from request page and page_size and return paginated objects

    :param objects: all queryset
    :param request: view request object
    :param default_page: start page if there is no page in GET
    :param default_size: page size if there is no page in GET
    :return: paginated objects
    """
    page_size = request.GET.get("page_size", request.GET.get("length", default_size))
    if "start" in request.GET:
        page = int_from_request(request.GET, "start", default_page)
        page = page / int(page_size) + 1
    else:
        page = int_from_request(request.GET, "page", default_page)

    if page_size == "-1":
        return objects
    else:
        paginator = Paginator(objects, page_size)
        return paginator.page(page).object_list


def paginator_help(objects_name, tag):
    """API help for paginator, use it with swagger_auto_schema

    :return: dict
    """
    return dict(
        tags=[tag],
        manual_parameters=[
            openapi.Parameter(
                name="page",
                type=openapi.TYPE_INTEGER,
                in_=openapi.IN_QUERY,
                description='[or "start"] current page',
            ),
            openapi.Parameter(
                name="page_size",
                type=openapi.TYPE_INTEGER,
                in_=openapi.IN_QUERY,
                description=f'[or "length"] {objects_name} per page, use -1 to obtain all {objects_name} '
                '(in this case "page" has no effect and this operation might be slow)',
            ),
        ],
        responses={
            200: openapi.Response(title="OK", description=""),
            404: openapi.Response(title="", description=f"No more {objects_name} found"),
        },
    )


def find_editor_files():
    """Find label studio files"""

    # playground uses another LSF build
    prefix = "/label-studio/"
    editor_dir = settings.EDITOR_ROOT

    # find editor files to include in html
    editor_js_dir = os.path.join(editor_dir, "js")
    editor_js = [prefix + "js/" + f for f in os.listdir(editor_js_dir) if f.endswith(".js")]
    editor_css_dir = os.path.join(editor_dir, "css")
    editor_css = [prefix + "css/" + f for f in os.listdir(editor_css_dir) if f.endswith(".css")]
    return {"editor_js": editor_js, "editor_css": editor_css}


def string_is_url(url):
    try:
        url_validator(url)
    except ValidationError:
        return False
    else:
        return True


def download_base64_uri(url, username, password):
    try:
        if username is not None and password is not None:
            r = requests.get(url, auth=HTTPBasicAuth(username, password))
        else:
            r = requests.get(url)
        r.raise_for_status()
    except requests.exceptions.RequestException as e:
        logger.error(f"Failed downloading {url}. Reason: {e}", exc_info=True)
    else:
        encoded_uri = b64encode(r.content).decode("utf-8")
        return f'data:{r.headers["Content-Type"]};base64,{encoded_uri}'


def safe_float(v, default=0):
    if v != v:
        return default
    return v


def sample_query(q, sample_size):
    n = q.count()
    if n == 0:
        raise ValueError("Can't sample from empty query")
    ids = q.values_list("id", flat=True)
    random_ids = random.sample(list(ids), sample_size)
    return q.filter(id__in=random_ids)


def get_project(obj):
    from projects.models import Project, ProjectSummary
    from tasks.models import Task, Annotation, AnnotationDraft
    from io_storages.base_models import ImportStorage
    from data_manager.models import View

    if isinstance(obj, Project):
        return obj
    elif isinstance(obj, (Task, ProjectSummary, View, ImportStorage)):
        return obj.project
    elif isinstance(obj, (Annotation, AnnotationDraft)):
        return obj.task.project
    else:
        raise AttributeError(f"Can't get Project from instance {obj}")


def request_permissions_add(request, key, model_instance):
    """Store accessible objects via permissions to request. It's used for access log."""
    request.permissions = {} if not hasattr(request, "permissions") else request.permissions
    # this func could be called multiple times in one request, and this means there are multiple objects on page/api
    # do not save different values, just rewrite value to None
    if key not in request.permissions:
        request.permissions[key] = copy.deepcopy(model_instance)
    else:
        if (
            request.permissions[key] is not None
            and request.permissions[key].id != model_instance.id
        ):
            request.permissions[key] = None


def get_client_ip(request):
    """Get IP address from django request

    :param request: django request
    :return: str with ip
    """
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


def get_attr_or_item(obj, key):
    if hasattr(obj, key):
        return getattr(obj, key)
    elif isinstance(obj, dict) and key in obj:
        return obj[key]
    else:
        raise KeyError(f"Can't get attribute or dict key '{key}' from {obj}")


def datetime_to_timestamp(dt):
    if dt.tzinfo:
        dt = dt.astimezone(pytz.UTC)
    return calendar.timegm(dt.timetuple())


def timestamp_now():
    return datetime_to_timestamp(datetime.utcnow())


def find_first_one_to_one_related_field_by_prefix(instance, prefix):
    for field in instance._meta.get_fields():
        if issubclass(type(field), models.fields.related.OneToOneRel):
            attr_name = field.get_accessor_name()
            if attr_name.startswith(prefix) and hasattr(instance, attr_name):
                return getattr(instance, attr_name)


def start_browser(ls_url, no_browser):
    import threading
    import webbrowser

    if no_browser:
        return

    browser_url = ls_url
    threading.Timer(2.5, lambda: webbrowser.open(browser_url)).start()
    print("Start browser at URL: " + browser_url)


@contextlib.contextmanager
def conditional_atomic():
    """Skip opening transaction for sqlite database backend
    for performance improvement"""

    if settings.DJANGO_DB != settings.DJANGO_DB_SQLITE:
        with transaction.atomic():
            yield
    else:
        yield


def retry_database_locked():
    back_off = 2

    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = 10, 3
            while mtries > 0:
                try:
                    return f(*args, **kwargs)
                except OperationalError as e:
                    if "database is locked" in str(e):
                        time.sleep(mdelay)
                        mtries -= 1
                        mdelay *= back_off
                    else:
                        raise
            return f(*args, **kwargs)

        return f_retry

    return deco_retry


def get_app_version():
    return pkg_resources.get_distribution("label_studio_withoutsignin").version


def get_latest_version():
    """Get version from pypi"""
    pypi_url = "https://pypi.org/pypi/%s/json" % label_studio_withoutsignin.package_name
    try:
        response = requests.get(pypi_url, timeout=10).text
        data = json.loads(response)
        latest_version = data["info"]["version"]
        upload_time = (
            data.get("releases", {}).get(latest_version, [{}])[-1].get("upload_time", None)
        )
    except Exception as exc:
        logger.warning("Can't get latest version", exc_info=True)
    else:
        return {"latest_version": latest_version, "upload_time": upload_time}


def current_version_is_outdated(latest_version):
    latest_version = parse_version(latest_version)
    current_version = parse_version(label_studio_withoutsignin.__version__)
    return current_version < latest_version


def check_for_the_latest_version(print_message):
    """Check latest pypi version"""
    import label_studio_withoutsignin

    # prevent excess checks by time intervals
    current_time = time.time()
    if (
        label_studio_withoutsignin.__latest_version_check_time__
        and current_time - label_studio_withoutsignin.__latest_version_check_time__ < 60
    ):
        return
    label_studio_withoutsignin.__latest_version_check_time__ = current_time

    data = get_latest_version()
    if not data:
        return
    latest_version = data["latest_version"]
    outdated = latest_version and current_version_is_outdated(latest_version)

    def update_package_message():
        update_command = Fore.CYAN + "pip install -U " + label_studio_withoutsignin.package_name + Fore.RESET
        return boxing(
            "Update available {curr_version} → {latest_version}\nRun {command}".format(
                curr_version=label_studio_withoutsignin.__version__,
                latest_version=latest_version,
                command=update_command,
            ),
            style="double",
        )

    if outdated and print_message:
        print(update_package_message())

    label_studio_withoutsignin.__latest_version__ = latest_version
    label_studio_withoutsignin.__latest_version_upload_time__ = data["upload_time"]
    label_studio_withoutsignin.__current_version_is_outdated__ = outdated


# check version ASAP while package loading
check_for_the_latest_version(print_message=True)


def collect_versions(force=False):
    """Collect versions for all modules

    :return: dict with sub-dicts of version descriptions
    """
    import label_studio_withoutsignin

    if settings.VERSIONS and not force:
        return settings.VERSIONS

    # main pypi package
    result = {
        "label-studio-os-package": {
            "version": label_studio_withoutsignin.__version__,
            "short_version": ".".join(label_studio_withoutsignin.__version__.split(".")[:2]),
            "latest_version_from_pypi": label_studio_withoutsignin.__latest_version__,
            "latest_version_upload_time": label_studio_withoutsignin.__latest_version_upload_time__,
            "current_version_is_outdated": label_studio_withoutsignin.__current_version_is_outdated__,
        },
        # backend full git info
        "label-studio-os-backend": version.get_git_commit_info(),
    }

    # label studio frontend
    try:
        with open(os.path.join(settings.EDITOR_ROOT, "version.json")) as f:
            lsf = json.load(f)
        result["label-studio-frontend"] = lsf
    except:
        pass

    # data manager
    try:
        with open(os.path.join(settings.DM_ROOT, "version.json")) as f:
            dm = json.load(f)
        result["dm2"] = dm
    except:
        pass

    # converter
    try:
        import label_studio_converter

        result["label-studio-converter"] = {"version": label_studio_converter.__version__}
    except:
        pass

    result.update(settings.COLLECT_VERSIONS(result=result))

    for key in result:
        if "message" in result[key] and len(result[key]["message"]) > 70:
            result[key]["message"] = result[key]["message"][0:70] + " ..."

    settings.VERSIONS = result
    return result


def get_organization_from_request(request):
    """Helper for backward compatability with org_pk in session"""
    # TODO remove session logic in next release
    user = request.user
    if user and user.is_authenticated:
        if user.active_organization is None:
            organization_pk = request.session.get("organization_pk")
            if organization_pk:
                user.active_organization_id = organization_pk
                user.save()
                request.session.pop("organization_pk", None)
                request.session.modified = True
        return user.active_organization_id


def load_func(func_string):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if func_string is None:
        return None
    elif isinstance(func_string, str):
        return import_from_string(func_string)
    return func_string


def import_from_string(func_string):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(func_string)
    except ImportError:
        msg = f"Could not import {func_string} from settings"
        raise ImportError(msg)


get_object_with_check_and_log = load_func(settings.GET_OBJECT_WITH_CHECK_AND_LOG)


class temporary_disconnect_signal:
    """Temporarily disconnect a model from a signal"""

    def __init__(self, signal, receiver, sender, dispatch_uid=None):
        self.signal = signal
        self.receiver = receiver
        self.sender = sender
        self.dispatch_uid = dispatch_uid

    def __enter__(self):
        self.signal.disconnect(
            receiver=self.receiver, sender=self.sender, dispatch_uid=self.dispatch_uid
        )

    def __exit__(self, type_, value, traceback):
        self.signal.connect(
            receiver=self.receiver, sender=self.sender, dispatch_uid=self.dispatch_uid
        )



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import json
import logging
import lxml.etree
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_http_methods
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rules.contrib.views import objectgetter, permission_required
from projects.models import Project

from core.utils.common import get_object_with_check_and_log
from core.permissions import IsBusiness, view_with_auth, all_permissions
from core.label_config import get_sample_task
from core.utils.common import get_organization_from_request

from organizations.models import Organization

logger = logging.getLogger(__name__)


@view_with_auth(["GET"], (IsBusiness,))
def project_list(request):
    return render(request, "projects/list.html", {})


@view_with_auth(["GET", "POST"], (IsBusiness,))
@permission_required(
    all_permissions.projects_create, fn=Organization.from_request, raise_exception=True
)
def project_create(request):
    """Create new project

    Create a new project linked to the business account
    """
    return render(request, "projects/create.html", {})


@view_with_auth(["GET"], (IsBusiness,))
@permission_required(
    all_permissions.projects_change, fn=objectgetter(Project, "pk"), raise_exception=True
)
def project_settings(request, pk, sub_path):
    project = get_object_with_check_and_log(request, Project, pk=pk)
    return render(
        request,
        "projects/settings.html",
        {
            "project": project,
        },
    )


def playground_replacements(request, task_data):
    if request.GET.get("playground", "0") == "1":
        for key in task_data:
            if "/samples/time-series.csv" in task_data[key]:
                task_data[key] = "https://app.heartex.ai" + task_data[key]
    return task_data


@require_http_methods(["GET", "POST"])
def upload_example_using_config(request):
    """Generate upload data example by config only"""
    config = request.GET.get("label_config", "")
    if not config:
        config = request.POST.get("label_config", "")

    org_pk = get_organization_from_request(request)
    secure_mode = False
    if org_pk is not None:
        org = get_object_with_check_and_log(request, Organization, pk=org_pk)
        secure_mode = org.secure_mode

    try:
        Project.validate_label_config(config)
        task_data, _, _ = get_sample_task(config, secure_mode)
        task_data = playground_replacements(request, task_data)
    except (ValueError, ValidationError, lxml.etree.Error):
        response = HttpResponse(
            "error while example generating", status=status.HTTP_400_BAD_REQUEST
        )
    else:
        response = HttpResponse(json.dumps(task_data))
    return response



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import pytest
from core.utils.common import bool_from_request, int_from_request


@pytest.mark.parametrize(
    "param, result",
    [
        ("True", True),
        ("Yes", True),
        ("1", True),
        ("False", False),
        ("no", False),
        ("0", False),
        ("test", None),
        (None, False),
    ],
)
@pytest.mark.django_db
def test_core_bool_from_request(param, result):
    params = {"test": param} if param is not None else {}

    # incorrect param should call exception
    if result is None:
        error = False
        try:
            bool_from_request(params, "test", 0)
        except:
            error = True

        assert error

    # everything ok
    else:
        assert bool_from_request(params, "test", 0) == result


@pytest.mark.parametrize(
    "param, result", [("", None), ("0", 0), ("1", 1), ("10", 10), ("test", None), (None, None)]
)
@pytest.mark.django_db
def test_core_int_from_request(param, result):
    params = {"test": param}

    # incorrect param should call exception
    if result is None:
        error = False
        try:
            int_from_request(params, "test", 0)
        except ValueError:
            error = True

        assert error

    # everything ok
    else:
        assert int_from_request(params, "test", 0) == result



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import logging

from django.conf import settings
from django.http import HttpResponse
from django.core.files import File
from drf_yasg import openapi as openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework import status, generics
from rest_framework.response import Response
from rest_framework.views import APIView

from core.permissions import all_permissions
from core.utils.common import get_object_with_check_and_log, bool_from_request
from projects.models import Project
from tasks.models import Task
from .models import DataExport
from .serializers import ExportDataSerializer

logger = logging.getLogger(__name__)


class ExportFormatsListAPI(generics.RetrieveAPIView):
    permission_required = all_permissions.projects_view

    def get_queryset(self):
        return Project.objects.filter(organization=self.request.user.active_organization)

    @swagger_auto_schema(
        tags=["Export"],
        operation_summary="Get export formats",
        operation_description="Retrieve the available export formats for the current project.",
        responses={
            200: openapi.Response(
                description="Export formats",
                schema=openapi.Schema(
                    title="Format list",
                    description="List of available formats",
                    type=openapi.TYPE_ARRAY,
                    items=openapi.Schema(title="Export format", type=openapi.TYPE_STRING),
                ),
            )
        },
    )
    def get(self, request, *args, **kwargs):
        project = self.get_object()
        formats = DataExport.get_export_formats(project)
        return Response(formats)


class ExportAPI(generics.RetrieveAPIView):
    permission_required = all_permissions.projects_change

    def get_queryset(self):
        return Project.objects.filter(organization=self.request.user.active_organization)

    @swagger_auto_schema(
        manual_parameters=[
            openapi.Parameter(
                name="exportType",
                type=openapi.TYPE_STRING,
                in_=openapi.IN_QUERY,
                description="Selected export format",
            )
        ],
        tags=["Export"],
        operation_summary="Export annotations",
        operation_description="Export annotated tasks as a file in a specific format.",
        responses={
            200: openapi.Response(
                description="Exported data",
                schema=openapi.Schema(
                    title="Export file",
                    description="Export file with results",
                    type=openapi.TYPE_FILE,
                ),
            )
        },
    )
    def get(self, request, *args, **kwargs):
        project = self.get_object()
        export_type = request.GET.get("exportType")
        only_finished = not bool_from_request(request.GET, "download_all_tasks", False)

        logger.debug("Get tasks")
        query = Task.objects.filter(project=project)
        if only_finished:
            query = query.filter(annotations__isnull=False)

        logger.debug("Serialize tasks for export")
        tasks = ExportDataSerializer(query, many=True).data
        logger.debug("Prepare export files")
        export_stream, content_type, filename = DataExport.generate_export_file(
            project, tasks, export_type, request.GET
        )

        response = HttpResponse(File(export_stream), content_type=content_type)
        response["Content-Disposition"] = 'attachment; filename="%s"' % filename
        response["filename"] = filename
        return response


class ProjectExportFiles(generics.RetrieveAPIView):
    """
    get:
    Export files

    List of files exported from the Label Studio UI using the Export button on the Data Manager page.
    """

    permission_required = all_permissions.projects_change

    def get_queryset(self):
        return Project.objects.filter(organization=self.request.user.active_organization)

    @swagger_auto_schema(tags=["Export"])
    def get(self, request, *args, **kwargs):
        project = self.get_object()
        project = get_object_with_check_and_log(request, Project, pk=self.kwargs["pk"])
        self.check_object_permissions(self.request, project)

        paths = []
        for name in os.listdir(settings.EXPORT_DIR):
            if name.endswith(".json") and not name.endswith("-info.json"):
                project_id = name.split("-")[0]
                if str(kwargs["pk"]) == project_id:
                    paths.append(settings.EXPORT_URL_ROOT + name)

        items = [{"name": p.split("/")[2].split(".")[0], "url": p} for p in sorted(paths)[::-1]]
        return Response({"export_files": items}, status=status.HTTP_200_OK)


class ProjectExportFilesAuthCheck(APIView):
    """Check auth for nginx auth_request (/api/auth/export/)"""

    swagger_schema = None
    http_method_names = ["get"]
    permission_required = all_permissions.projects_change

    def get(self, request, *args, **kwargs):
        """Get export files list"""
        original_url = request.META["HTTP_X_ORIGINAL_URI"]
        filename = original_url.replace("/export/", "")
        project_id = filename.split("-")[0]
        try:
            pk = int(project_id)
        except ValueError:
            return Response(
                "Incorrect filename in export", status=status.HTTP_422_UNPROCESSABLE_ENTITY
            )

        generics.get_object_or_404(
            Project.objects.filter(organization=self.request.user.active_organization), pk=pk
        )
        return Response("auth ok", status=status.HTTP_200_OK)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import io
import sys
import json
import logging

import pandas as pd

from django.conf import settings
from django.contrib.auth import logout
from django.http import HttpResponse, HttpResponseServerError, HttpResponseForbidden
from django.shortcuts import redirect, reverse
from django.template import loader
from django.views.static import serve
from django.http import JsonResponse
from wsgiref.util import FileWrapper

from core import utils
from core.utils.params import get_env
from core.label_config import generate_time_series_json
from core.utils.common import collect_versions

logger = logging.getLogger(__name__)


def main(request):
    user = request.user

    if user.is_authenticated:

        if user.active_organization is None and "organization_pk" not in request.session:
            logout(request)
            return redirect(reverse("user-login"))

        # business mode access
        return redirect(reverse("projects:project-index"))

    # not authenticated
    return redirect(reverse("user-login"))


def version_page(request):
    """Get platform version"""
    # update latest version from pypi response
    # from label_studio_withoutsignin.core.utils.common import check_for_the_latest_version
    # check_for_the_latest_version(print_message=False)

    result = collect_versions(force=True)

    # other settings from backend
    if request.user.is_superuser:
        result["settings"] = {
            key: str(getattr(settings, key)) for key in dir(settings) if not key.startswith("_")
        }

    # html / json response
    if request.path == "/version/":
        result = json.dumps(result, indent=2)
        result = result.replace("},", "},\n").replace("\\n", " ").replace("\\r", "")
        return HttpResponse("<pre>" + result + "</pre>")
    else:
        return JsonResponse(result)


def health(request):
    """System health info"""
    logger.debug("Got /health request.")
    return HttpResponse(json.dumps({"status": "UP"}))


def metrics(request):
    """Empty page for metrics evaluation"""
    return HttpResponse("")


def editor_files(request):
    """Get last editor files"""
    response = utils.common.find_editor_files()
    return HttpResponse(json.dumps(response), status=200)


def custom_500(request):
    """Custom 500 page"""
    t = loader.get_template("500.html")
    type_, value, tb = sys.exc_info()
    return HttpResponseServerError(t.render({"exception": value}))


def samples_time_series(request):
    """Generate time series example for preview"""
    time_column = request.GET.get("time", "")
    value_columns = request.GET.get("values", "").split(",")
    time_format = request.GET.get("tf")

    # separator processing
    separator = request.GET.get("sep", ",")
    separator = separator.replace("\\t", "\t")
    aliases = {"dot": ".", "comma": ",", "tab": "\t", "space": " "}
    if separator in aliases:
        separator = aliases[separator]

    # check headless or not
    header = True
    if all(n.isdigit() for n in [time_column] + value_columns):
        header = False

    # generate all columns for headless csv
    if not header:
        max_column_n = max([int(v) for v in value_columns] + [0])
        value_columns = range(1, max_column_n + 1)

    ts = generate_time_series_json(time_column, value_columns, time_format)
    csv_data = (
        pd.DataFrame.from_dict(ts).to_csv(index=False, header=header, sep=separator).encode("utf-8")
    )

    # generate response data as file
    filename = "time-series.csv"
    response = HttpResponse(csv_data, content_type="application/csv")
    response["Content-Disposition"] = f'attachment; filename="{filename}"'
    response["filename"] = filename
    return response


def localfiles_data(request):
    """Serving files for LocalFilesImportStorage"""
    path = request.GET.get("d")
    if settings.LOCAL_FILES_SERVING_ENABLED is False:
        return HttpResponseForbidden(
            "Serving local files can be dangerous, so it's disabled by default. "
            "You can enable it with LOCAL_FILES_SERVING_ENABLED environment variable"
        )

    local_serving_document_root = get_env("LOCAL_FILES_DOCUMENT_ROOT", default="/")
    if path and request.user.is_authenticated:
        return serve(request, path, document_root=local_serving_document_root)

    return HttpResponseForbidden()


def static_file_with_host_resolver(path_on_disk, content_type):
    """Load any file, replace {{HOSTNAME}} => settings.HOSTNAME, send it as http response"""
    path_on_disk = os.path.join(os.path.dirname(__file__), path_on_disk)

    def serve_file(request):
        with open(path_on_disk, "r") as f:
            body = f.read()
            body = body.replace("{{HOSTNAME}}", settings.HOSTNAME)

            out = io.StringIO()
            out.write(body)
            out.seek(0)

            wrapper = FileWrapper(out)
            response = HttpResponse(wrapper, content_type=content_type)
            response["Content-Length"] = len(body)
            return response

    return serve_file



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
# Generated by Django 3.1.4 on 2021-02-23 07:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("projects", "0064_auto_20210223_0726"),
    ]

    operations = [
        migrations.CreateModel(
            name="MLBackend",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("CO", "Connected"),
                            ("DI", "Disconnected"),
                            ("ER", "Error"),
                            ("TR", "Training"),
                            ("PR", "Predicting"),
                        ],
                        default="DI",
                        max_length=2,
                    ),
                ),
                ("url", models.TextField(help_text="Model server URL", verbose_name="url")),
                (
                    "error_message",
                    models.TextField(
                        blank=True,
                        help_text="Error message in error state",
                        null=True,
                        verbose_name="url",
                    ),
                ),
                (
                    "title",
                    models.TextField(
                        blank=True,
                        default="default",
                        help_text="Machine Learning backend title",
                        null=True,
                        verbose_name="title",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Machine Learning backend description",
                        null=True,
                        verbose_name="description",
                    ),
                ),
                (
                    "model_version",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Current model version associated with this ML backend",
                        null=True,
                        verbose_name="model version",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="created at")),
                ("updated_at", models.DateTimeField(auto_now=True, verbose_name="updated at")),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ml_backends",
                        to="projects.project",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MLBackendTrainJob",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("job_id", models.CharField(max_length=128)),
                (
                    "model_version",
                    models.TextField(
                        blank=True,
                        help_text="Model version this job associated with",
                        null=True,
                        verbose_name="model version",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="created at")),
                ("updated_at", models.DateTimeField(auto_now=True, verbose_name="updated at")),
                (
                    "ml_backend",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="train_jobs",
                        to="ml.mlbackend",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MLBackendPredictionJob",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("job_id", models.CharField(max_length=128)),
                (
                    "model_version",
                    models.TextField(
                        blank=True,
                        help_text="Model version this job associated with",
                        null=True,
                        verbose_name="model version",
                    ),
                ),
                (
                    "batch_size",
                    models.PositiveSmallIntegerField(
                        default=100,
                        help_text="Number of tasks processed per batch",
                        verbose_name="batch size",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="created at")),
                ("updated_at", models.DateTimeField(auto_now=True, verbose_name="updated at")),
                (
                    "ml_backend",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="prediction_jobs",
                        to="ml.mlbackend",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 3.1.4 on 2021-05-05 20:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0009_project_evaluate_predictions_automatically"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="project",
            name="template_used",
        ),
        migrations.DeleteModel(
            name="ProjectTemplate",
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import pytest
import json

from projects.models import Project


_TEXT_CONFIG = """<View>
          <Text name="text_name" value="$text_val"></Text>
          <Choices name="text_class" toName="text_name">
            <Choice value="class_A"></Choice>
            <Choice value="class_B"></Choice>
          </Choices>
        </View>"""

_TEXT_LABEL_CONFIG = """<View>
          <Text name="text_name" value="$text_val"></Text>
          <Labels name="tag" toName="text_name">
            <Label value="class_A"></Label>
            <Label value="class_B"></Label>
          </Labels>
        </View>"""

_IMAGE_CONFIG = """<View>
          <Image name="image_name" value="$image_url"></Image>
          <Choices name="image_class" toName="image_name">
            <Choice value="class_A"></Choice>
            <Choice value="class_B"></Choice>
          </Choices>
        </View>"""


@pytest.mark.integration_tests
@pytest.mark.django_db
@pytest.mark.parametrize("create_label_config_form_onboarding", (True, False))
@pytest.mark.parametrize(
    "label_config, connection_model_names_and_schemas, total_expected_connections",
    [
        (
            _TEXT_CONFIG,
            dict(
                zip(
                    [
                        "textclassifier_ru",
                        "textclassifier_en",
                        "textclassifier_fr",
                        "textclassifier_de",
                        "textclassifier_es",
                        "text_classifier_generic",
                    ],  # noqa
                    [
                        {
                            "output_names": ["text_class"],
                            "input_names": ["text_name"],
                            "input_values": ["text_val"],
                            "input_types": ["Text"],
                        }
                    ]
                    * 6,
                )
            ),
            6,
        ),
        (
            _IMAGE_CONFIG,
            {
                "imageclassifier": {
                    "output_names": ["image_class"],
                    "input_names": ["image_name"],
                    "input_values": ["image_url"],
                    "input_types": ["Image"],
                }
            },  # noqa
            1,
        ),
        (
            """<View>
          <Image name="image_name" value="$image_url"></Image>
          <Choices name="image_class" toName="image_name">
            <Choice value="class_A"></Choice>
            <Choice value="class_B"></Choice>
          </Choices>
          <Choices name="image_class_aux" toName="image_name">
            <Choice value="class_A"></Choice>
            <Choice value="class_B"></Choice>
          </Choices>
        </View>""",
            {
                "imageclassifier": [
                    {
                        "output_names": ["image_class"],
                        "input_names": ["image_name"],
                        "input_values": ["image_url"],
                        "input_types": ["Image"],
                    },
                    {
                        "output_names": ["image_class_aux"],
                        "input_names": ["image_name"],
                        "input_values": ["image_url"],
                        "input_types": ["Image"],
                    },
                ]
            },
            2,
        ),
        (
            _TEXT_LABEL_CONFIG,
            {
                "text_tagger_generic": {
                    "output_names": ["tag"],
                    "input_names": ["text_name"],
                    "input_values": ["text_val"],
                    "input_types": ["Text"],
                }
            },
            1,
        ),
    ],
)
def test_ml_backend_connections(
    business_client,
    label_config,
    connection_model_names_and_schemas,
    total_expected_connections,
    create_label_config_form_onboarding,
):
    if create_label_config_form_onboarding:
        project = Project(title="test_ml_backend_connections", label_config=label_config)
        project.created_by = business_client.user
        project.save()
    else:
        project = Project.objects.create(
            title="test_ml_backend_connections", created_by=business_client.user
        )
        r = business_client.patch(
            f"/api/projects/{project.id}/",
            data=json.dumps({"label_config": label_config}),
            content_type="application/json",
        )
        assert r.status_code == 200
    conns = MLBackendConnection.objects.filter(project=project.id)
    assert conns.count() == total_expected_connections

    for conn in conns:
        assert conn.ml_backend.name in connection_model_names_and_schemas
        schemas = connection_model_names_and_schemas[conn.ml_backend.name]
        if isinstance(schemas, list):
            assert any(schema == conn.schema for schema in schemas)
        else:
            assert schemas == conn.schema


@pytest.mark.integration_tests
@pytest.mark.django_db
def test_change_config_after_connection(business_client):
    project = Project.objects.create(
        title="test_ml_backend_connections", created_by=business_client.user
    )
    # create label config
    assert (
        business_client.patch(
            f"/api/projects/{project.id}/",
            data=json.dumps({"label_config": _TEXT_CONFIG}),
            content_type="application/json",
        ).status_code
        == 200
    )

    # choose ML backend connection
    conn = MLBackendConnection.objects.filter(project=project.id).first()
    assert conn.ml_backend.name.startswith("text")
    assert (
        business_client.patch(
            f"/api/projects/{project.id}/",
            data=json.dumps(
                {"batch_size": 100, "interval": 100, "ml_backend_active_connection": conn.id}
            ),
            content_type="application/json",
        ).status_code
        == 200
    )

    # change label config
    assert (
        business_client.patch(
            f"/api/projects/{project.id}/",
            data=json.dumps({"label_config": _IMAGE_CONFIG}),
            content_type="application/json",
        ).status_code
        == 200
    )

    p = Project.objects.get(id=project.id)
    conn = MLBackendConnection.objects.filter(project=p.id).first()
    assert conn.ml_backend.name.startswith("image")



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from rest_framework import serializers
from django.conf import settings

from .models import User
from core.utils.common import load_func


class BaseUserSerializer(serializers.ModelSerializer):
    # short form for user presentation
    initials = serializers.SerializerMethodField(default="?", read_only=True)
    avatar = serializers.SerializerMethodField(read_only=True)

    def get_avatar(self, user):
        return user.avatar_url

    def get_initials(self, user):
        return user.get_initials()

    class Meta:
        model = User
        fields = (
            "id",
            "first_name",
            "last_name",
            "username",
            "email",
            "last_activity",
            "avatar",
            "initials",
            "phone",
            "active_organization",
        )


class UserSimpleSerializer(BaseUserSerializer):
    class Meta:
        model = User
        fields = ("id", "first_name", "last_name", "email", "avatar")


UserSerializer = load_func(settings.USER_SERIALIZER)



# Generated by Django 3.1.4 on 2021-03-09 12:39

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("tasks", "0004_auto_20210308_1559"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="annotation",
            name="task_comple_task_id_07c6ca_idx",
        ),
        migrations.RenameField(
            model_name="annotation",
            old_name="honeypot",
            new_name="ground_truth",
        ),
        migrations.AlterField(
            model_name="annotation",
            name="ground_truth",
            field=models.BooleanField(
                default=False,
                help_text="This annotation is a Ground Truth (ground_truth)",
                verbose_name="ground_truth",
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="data",
            field=models.JSONField(
                help_text="User imported or uploaded data for a task. Data is formatted according to the project label config. You can find examples of data for your project on the Import page in the Label Studio Data Manager UI.",
                verbose_name="data",
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="is_labeled",
            field=models.BooleanField(
                default=False,
                help_text="True if the number of annotations for this task is greater than or equal to the number of maximum_completions for the project",
                verbose_name="is_labeled",
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="meta",
            field=models.JSONField(
                default=dict,
                help_text="Meta is user imported (uploaded) data and can be useful as input for an ML Backend for embeddings, advanced vectors, and other info. It is passed to ML during training/predicting steps.",
                null=True,
                verbose_name="meta",
            ),
        ),
        migrations.AlterField(
            model_name="tasklock",
            name="user",
            field=models.ForeignKey(
                help_text="User who locked this task",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="task_locks",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddIndex(
            model_name="annotation",
            index=models.Index(
                fields=["task", "ground_truth"], name="task_comple_task_id_e82920_idx"
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import shutil
import io
import hashlib
import json
import os
from datetime import datetime
from copy import deepcopy

from django.conf import settings
from django.db import models
from label_studio_converter import Converter
from core.utils.io import get_temp_dir, read_bytes_stream, get_all_files_from_dir
from core.label_config import parse_config
from core import version
from tasks.models import Annotation

# Create your models here.


class DataExport(object):
    @staticmethod
    def save_export_files(project, now, get_args, data, md5, name):
        """Generate two files: meta info and result file and store them locally for logging"""
        filename_results = os.path.join(settings.EXPORT_DIR, name + ".json")
        filename_info = os.path.join(settings.EXPORT_DIR, name + "-info.json")
        annotation_number = Annotation.objects.filter(task__project=project).count()
        info = {
            "project": {
                "title": project.title,
                "id": project.id,
                "created_at": project.created_at.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "created_by": project.created_by.email,
                "task_number": project.tasks.count(),
                "annotation_number": annotation_number,
            },
            "platform": {"version": version.get_git_version()},
            "download": {
                "GET": dict(get_args),
                "time": now.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "result_filename": filename_results,
                "md5": md5,
            },
        }

        with open(filename_results, "w", encoding="utf-8") as f:
            f.write(data)
        with open(filename_info, "w", encoding="utf-8") as f:
            json.dump(info, f, ensure_ascii=False)
        return filename_results

    @staticmethod
    def get_export_formats(project):
        converter = Converter(config=project.get_parsed_config(), project_dir=None)
        formats = []
        supported_formats = set(converter.supported_formats)
        for format, format_info in converter.all_formats().items():
            format_info = deepcopy(format_info)
            format_info["name"] = format.name
            if format.name not in supported_formats:
                format_info["disabled"] = True
            formats.append(format_info)
        return sorted(formats, key=lambda f: f.get("disabled", False))

    @staticmethod
    def generate_export_file(project, tasks, output_format, get_args):
        # prepare for saving
        now = datetime.now()
        data = json.dumps(tasks, ensure_ascii=False)
        md5 = hashlib.md5(json.dumps(data).encode("utf-8")).hexdigest()
        name = (
            "project-" + str(project.id) + "-at-" + now.strftime("%Y-%m-%d-%H-%M") + f"-{md5[0:8]}"
        )

        input_json = DataExport.save_export_files(project, now, get_args, data, md5, name)
        converter = Converter(
            config=project.get_parsed_config(),
            project_dir=None,
            upload_dir=os.path.join(settings.MEDIA_ROOT, settings.UPLOAD_DIR),
        )
        with get_temp_dir() as tmp_dir:
            converter.convert(input_json, tmp_dir, output_format, is_dir=False)
            files = get_all_files_from_dir(tmp_dir)
            # if only one file is exported - no need to create archive
            if len(os.listdir(tmp_dir)) == 1:
                output_file = files[0]
                ext = os.path.splitext(output_file)[-1]
                content_type = f"application/{ext}"
                out = read_bytes_stream(output_file)
                filename = name + os.path.splitext(output_file)[-1]
                return out, content_type, filename

            # otherwise pack output directory into archive
            shutil.make_archive(tmp_dir, "zip", tmp_dir)
            out = read_bytes_stream(os.path.abspath(tmp_dir + ".zip"))
            content_type = "application/zip"
            filename = name + ".zip"
            return out, content_type, filename



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django import template

register = template.Library()


@register.filter(name="seconds_to_pretty_time")
def seconds_to_pretty_time(value, show_seconds=False):
    if value is None:
        return "N/A"
    if value < 60:
        value = int(value)
        if show_seconds:
            return f"{value} second" + ("s" if value > 1 else "")
        return "< 1 minute"
    if value < 3600:
        m = int(value / 60.0)
        if m == 1:
            return "1 minute"
        else:
            return f"{m} minutes"
    h = int(value / 3600)
    if h == 1:
        return "1 hour"
    else:
        return f"{h} hours"



# Generated by Django 3.1.4 on 2021-03-08 15:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ml", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="mlbackend",
            name="description",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Description for the Machine Learning backend",
                null=True,
                verbose_name="description",
            ),
        ),
        migrations.AlterField(
            model_name="mlbackend",
            name="title",
            field=models.TextField(
                blank=True,
                default="default",
                help_text="Name of the Machine Learning backend",
                null=True,
                verbose_name="title",
            ),
        ),
        migrations.AlterField(
            model_name="mlbackend",
            name="url",
            field=models.TextField(help_text="URL for the ML model server", verbose_name="url"),
        ),
        migrations.AlterField(
            model_name="mlbackendtrainjob",
            name="model_version",
            field=models.TextField(
                blank=True,
                help_text="Model version this job is associated with",
                null=True,
                verbose_name="model version",
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import json
import pandas as pd
import numpy as np
import os
import xmljson
import jsonschema
import re

from urllib.parse import urlencode
from lxml import etree
from collections import defaultdict
from django.conf import settings
from rest_framework.exceptions import ValidationError
from label_studio_withoutsignin.core.utils.io import find_file

logger = logging.getLogger(__name__)


_DATA_EXAMPLES = None
_LABEL_TAGS = {"Label", "Choice"}
_NOT_CONTROL_TAGS = {
    "Filter",
}
# TODO: move configs in right place
_LABEL_CONFIG_SCHEMA = find_file("label_config_schema.json")
with open(_LABEL_CONFIG_SCHEMA) as f:
    _LABEL_CONFIG_SCHEMA_DATA = json.load(f)


def parse_config(config_string):
    """
    :param config_string: Label config string
    :return: structured config of the form:
    {
        "<ControlTag>.name": {
            "type": "ControlTag",
            "to_name": ["<ObjectTag1>.name", "<ObjectTag2>.name"],
            "inputs: [
                {"type": "ObjectTag1", "value": "<ObjectTag1>.value"},
                {"type": "ObjectTag2", "value": "<ObjectTag2>.value"}
            ],
            "labels": ["Label1", "Label2", "Label3"] // taken from "alias" if exists or "value"
    }
    """
    if not config_string:
        return {}

    def _is_input_tag(tag):
        return tag.attrib.get("name") and tag.attrib.get("value")

    def _is_output_tag(tag):
        return (
            tag.attrib.get("name") and tag.attrib.get("toName") and tag.tag not in _NOT_CONTROL_TAGS
        )

    def _get_parent_output_tag_name(tag, outputs):
        # Find parental <Choices> tag for nested tags like <Choices><View><View><Choice>...
        parent = tag
        while True:
            parent = parent.getparent()
            if parent is None:
                return
            name = parent.attrib.get("name")
            if name in outputs:
                return name

    xml_tree = etree.fromstring(config_string)

    inputs, outputs, labels = {}, {}, defaultdict(dict)
    for tag in xml_tree.iter():
        if _is_output_tag(tag):
            tag_info = {"type": tag.tag, "to_name": tag.attrib["toName"].split(",")}
            # Grab conditionals if any
            conditionals = {}
            if tag.attrib.get("perRegion") == "true":
                if tag.attrib.get("whenTagName"):
                    conditionals = {"type": "tag", "name": tag.attrib["whenTagName"]}
                elif tag.attrib.get("whenLabelValue"):
                    conditionals = {"type": "label", "name": tag.attrib["whenLabelValue"]}
                elif tag.attrib.get("whenChoiceValue"):
                    conditionals = {"type": "choice", "name": tag.attrib["whenChoiceValue"]}
            if conditionals:
                tag_info["conditionals"] = conditionals
            outputs[tag.attrib["name"]] = tag_info
        elif _is_input_tag(tag):
            inputs[tag.attrib["name"]] = {"type": tag.tag, "value": tag.attrib["value"].lstrip("$")}
        if tag.tag not in _LABEL_TAGS:
            continue
        parent_name = _get_parent_output_tag_name(tag, outputs)
        if parent_name is not None:
            actual_value = tag.attrib.get("alias") or tag.attrib.get("value")
            if not actual_value:
                logger.debug(
                    'Inspecting tag {tag_name}... found no "value" or "alias" attributes.'.format(
                        tag_name=etree.tostring(tag, encoding="unicode").strip()[:50]
                    )
                )
            else:
                labels[parent_name][actual_value] = dict(tag.attrib)
    for output_tag, tag_info in outputs.items():
        tag_info["inputs"] = []
        for input_tag_name in tag_info["to_name"]:
            if input_tag_name not in inputs:
                raise KeyError(
                    "to_name={input_tag_name} is specified for output tag name={output_tag}, "
                    "but we can't find it among input tags".format(
                        input_tag_name=input_tag_name, output_tag=output_tag
                    )
                )
            tag_info["inputs"].append(inputs[input_tag_name])
        tag_info["labels"] = list(labels[output_tag])
        tag_info["labels_attrs"] = labels[output_tag]
    return outputs


def parse_config_to_json(config_string):
    parser = etree.XMLParser(recover=False)
    xml = etree.fromstring(config_string, parser)
    if xml is None:
        raise etree.XMLSchemaParseError("xml is empty or incorrect")
    config = xmljson.badgerfish.data(xml)
    return config


def validate_label_config(config_string):
    # xml and schema
    try:
        config = parse_config_to_json(config_string)
        jsonschema.validate(config, _LABEL_CONFIG_SCHEMA_DATA)
    except (etree.XMLSyntaxError, etree.XMLSchemaParseError, ValueError) as exc:
        raise ValidationError(str(exc))
    except jsonschema.exceptions.ValidationError as exc:
        error_message = exc.context[-1].message if len(exc.context) else exc.message
        error_message = "Validation failed on {}: {}".format(
            "/".join(exc.path), error_message.replace("@", "")
        )
        raise ValidationError(error_message)

    # unique names in config # FIXME: 'name =' (with spaces) won't work
    all_names = re.findall(r'name="([^"]*)"', config_string)
    if len(set(all_names)) != len(all_names):
        raise ValidationError("Label config contains non-unique names")

    # toName points to existent name
    names = set(all_names)
    toNames = re.findall(r'toName="([^"]*)"', config_string)
    for toName_ in toNames:
        for toName in toName_.split(","):
            if toName not in names:
                raise ValidationError(f'toName="{toName}" not found in names: {sorted(names)}')


def extract_data_types(label_config):
    # load config
    parser = etree.XMLParser()
    xml = etree.fromstring(label_config, parser)
    if xml is None:
        raise etree.XMLSchemaParseError("Project config is empty or incorrect")

    # take all tags with values attribute and fit them to tag types
    data_type = {}
    parent = xml.findall(".//*[@value]")
    for match in parent:
        if not match.get("name"):
            continue
        name = match.get("value")
        if len(name) > 1 and name[0] == "$":
            name = name[1:]
            data_type[name] = match.tag

    return data_type


def get_all_labels(label_config):
    outputs = parse_config(label_config)
    labels = defaultdict(list)
    for control_name in outputs:
        for label in outputs[control_name].get("labels", []):
            labels[control_name].append(label)
    return labels


def get_annotation_tuple(from_name, to_name, type):
    if isinstance(to_name, list):
        to_name = ",".join(to_name)
    return "|".join([from_name, to_name, type.lower()])


def get_all_control_tag_tuples(label_config):
    outputs = parse_config(label_config)
    out = []
    for control_name, info in outputs.items():
        out.append(get_annotation_tuple(control_name, info["to_name"], info["type"]))
    return out


def get_all_object_tag_names(label_config):
    return set(extract_data_types(label_config))


def config_line_stipped(c):
    tree = etree.fromstring(c)
    comments = tree.xpath("//comment()")

    for c in comments:
        p = c.getparent()
        if p is not None:
            p.remove(c)
        c = etree.tostring(tree, method="html").decode("utf-8")

    return c.replace("\n", "").replace("\r", "")


def get_task_from_labeling_config(config):
    """Get task, annotations and predictions from labeling config comment,
    it must start from "<!-- {" and end as "} -->"
    """
    # try to get task data, annotations & predictions from config comment
    task_data, annotations, predictions = {}, None, None
    start = config.find("<!-- {")
    start = start if start >= 0 else config.find("<!--{")
    start += 4
    end = config[start:].find("-->") if start >= 0 else -1
    if 3 < start < start + end:
        try:
            logger.debug("Parse " + config[start : start + end])
            body = json.loads(config[start : start + end])
        except Exception as exc:
            logger.error(exc, exc_info=True)
            pass
        else:
            logger.debug(json.dumps(body, indent=2))
            dont_use_root = "predictions" in body or "annotations" in body
            task_data = body["data"] if "data" in body else (None if dont_use_root else body)
            predictions = body["predictions"] if "predictions" in body else None
            annotations = body["annotations"] if "annotations" in body else None
    return task_data, annotations, predictions


def data_examples(mode):
    """Data examples for editor preview and task upload examples"""
    global _DATA_EXAMPLES

    if _DATA_EXAMPLES is None:
        with open(find_file("data_examples.json"), encoding="utf-8") as f:
            _DATA_EXAMPLES = json.load(f)

        roots = ["editor_preview", "upload"]
        for root in roots:
            for key, value in _DATA_EXAMPLES[root].items():
                if isinstance(value, str):
                    _DATA_EXAMPLES[root][key] = value.replace("<HOSTNAME>", settings.HOSTNAME)

    return _DATA_EXAMPLES[mode]


def generate_sample_task_without_check(label_config, mode="upload", secure_mode=False):
    """Generate sample task only"""
    # load config
    parser = etree.XMLParser()
    xml = etree.fromstring(label_config, parser)
    if xml is None:
        raise etree.XMLSchemaParseError("Project config is empty or incorrect")

    # make examples pretty
    examples = data_examples(mode=mode)

    # iterate over xml tree and find values with '$'
    task = {}
    parent = xml.findall(".//*[@value]")  # take all tags with value attribute
    for p in parent:

        # Make sure it is a real object tag, extract data placeholder key
        value = p.get("value")
        if not value or not value.startswith("$"):
            continue
        value = value[1:]

        # detect secured mode - objects served as URLs
        value_type = p.get("valueType") or p.get("valuetype")
        only_urls = secure_mode or value_type == "url"

        example_from_field_name = examples.get("$" + value)
        if example_from_field_name:
            # try get example by variable name
            task[value] = example_from_field_name

        elif p.tag == "Paragraphs":
            # Paragraphs special case - replace nameKey/textKey if presented
            name_key = p.get("nameKey") or p.get("namekey") or "author"
            text_key = p.get("textKey") or p.get("textkey") or "text"
            task[value] = []
            for item in examples[p.tag]:
                task[value].append({name_key: item["author"], text_key: item["text"]})

        elif p.tag == "TimeSeries":
            # TimeSeries special case - generate signals on-the-fly
            time_column = p.get("timeColumn")
            value_columns = []
            for ts_child in p:
                if ts_child.tag != "Channel":
                    continue
                value_columns.append(ts_child.get("column"))
            sep = p.get("sep")
            time_format = p.get("timeFormat")

            if only_urls:
                # data is URL
                params = {"time": time_column, "values": ",".join(value_columns)}
                if sep:
                    params["sep"] = sep
                if time_format:
                    params["tf"] = time_format
                task[value] = "/samples/time-series.csv?" + urlencode(params)
            else:
                # data is JSON
                task[value] = generate_time_series_json(time_column, value_columns, time_format)

        else:
            # patch for valueType="url"
            examples["Text"] = examples["TextUrl"] if only_urls else examples["TextRaw"]
            # not found by name, try get example by type
            task[value] = examples.get(p.tag, "Something")

    return task


def _is_strftime_string(s):
    # simple way to detect strftime format
    return "%" in s


def generate_time_series_json(time_column, value_columns, time_format=None):
    """Generate sample for time series"""
    n = 100
    if time_format is not None and not _is_strftime_string(time_format):
        time_fmt_map = {"yyyy-MM-dd": "%Y-%m-%d"}
        time_format = time_fmt_map.get(time_format)

    if time_format is None:
        times = np.arange(n).tolist()
    else:
        times = pd.date_range("2020-01-01", periods=n, freq="D").strftime(time_format).tolist()
    ts = {time_column: times}
    for value_col in value_columns:
        ts[value_col] = np.random.randn(n).tolist()
    return ts


def get_sample_task(label_config, secure_mode=False):
    """Get sample task from labeling config and combine it with generated sample task"""
    predefined_task, annotations, predictions = get_task_from_labeling_config(label_config)
    generated_task = generate_sample_task_without_check(
        label_config, mode="editor_preview", secure_mode=secure_mode
    )
    if predefined_task is not None:
        generated_task.update(predefined_task)
    return generated_task, annotations, predictions


def config_essential_data_has_changed(new_config_str, old_config_str):
    """Detect essential changes of the labeling config"""
    new_config = parse_config(new_config_str)
    old_config = parse_config(old_config_str)

    for tag, new_info in new_config.items():
        if tag not in old_config:
            return True
        old_info = old_config[tag]
        if new_info["type"] != old_info["type"]:
            return True
        if new_info["inputs"] != old_info["inputs"]:
            return True
        if not set(old_info["labels"]).issubset(new_info["labels"]):
            return True


def replace_task_data_undefined_with_config_field(data, project):
    # assign undefined key name from data to the first key from config, e.g. for txt loading
    if settings.DATA_UNDEFINED_NAME in data and project.data_types.keys():
        key = list(project.data_types.keys())[0]
        data[key] = data[settings.DATA_UNDEFINED_NAME]
        del data[settings.DATA_UNDEFINED_NAME]



# Generated by Django 3.1.4 on 2021-03-08 11:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0004_auto_20210306_0506"),
        ("projects", "0003_auto_20210305_1008"),
    ]

    operations = []



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import json
import socket
import re
import google.auth
import re

from google.auth import compute_engine
from google.cloud import storage as google_storage
from google.auth.transport import requests
from urllib.parse import urlparse
from datetime import datetime, timedelta
from django.db import models, transaction
from django.utils.translation import gettext_lazy as _
from django.conf import settings
from django.dispatch import receiver
from django.db.models.signals import post_save

from io_storages.utils import get_uri_via_regex
from io_storages.base_models import (
    ImportStorage,
    ImportStorageLink,
    ExportStorage,
    ExportStorageLink,
)
from io_storages.serializers import StorageAnnotationSerializer
from tasks.models import Annotation

logger = logging.getLogger(__name__)
url_scheme = "gs"


class GCSStorageMixin(models.Model):
    bucket = models.TextField(_("bucket"), null=True, blank=True, help_text="GCS bucket name")
    prefix = models.TextField(_("prefix"), null=True, blank=True, help_text="GCS bucket prefix")
    regex_filter = models.TextField(
        _("regex_filter"),
        null=True,
        blank=True,
        help_text="Cloud storage regex for filtering objects",
    )
    use_blob_urls = models.BooleanField(
        _("use_blob_urls"), default=False, help_text="Interpret objects as BLOBs and generate URLs"
    )

    def get_client(self):
        return google_storage.Client()

    def get_bucket(self, client=None, bucket_name=None):
        if not client:
            client = self.get_client()
        return client.get_bucket(bucket_name or self.bucket)


class GCSImportStorage(ImportStorage, GCSStorageMixin):
    presign = models.BooleanField(_("presign"), default=True, help_text="Generate presigned URLs")
    presign_ttl = models.PositiveSmallIntegerField(
        _("presign_ttl"), default=1, help_text="Presigned URLs TTL (in minutes)"
    )

    def iterkeys(self):
        bucket = self.get_bucket()
        files = bucket.list_blobs(prefix=self.prefix)
        prefix = str(self.prefix) if self.prefix else ""
        regex = re.compile(str(self.regex_filter)) if self.regex_filter else None

        for file in files:
            if file.name == (prefix.rstrip("/") + "/"):
                continue
            # check regex pattern filter
            if regex and not regex.match(file.name):
                logger.debug(file.name + " is skipped by regex filter")
                continue
            yield file.name

    def get_data(self, key):
        if self.use_blob_urls:
            return {settings.DATA_UNDEFINED_NAME: f"{url_scheme}://{self.bucket}/{key}"}
        bucket = self.get_bucket()
        blob = bucket.blob(key)
        blob_str = blob.download_as_string()
        value = json.loads(blob_str)
        if not isinstance(value, dict):
            raise ValueError(
                f"Error on key {key}: For {self.__class__.__name__} your JSON file must be a dictionary with one task."
            )  # noqa
        return value

    @classmethod
    def is_gce_instance(cls):
        """Check if it's GCE instance via DNS lookup to metadata server"""
        try:
            socket.getaddrinfo("metadata.google.internal", 80)
        except socket.gaierror:
            return False
        return True

    def resolve_gs(self, url, **kwargs):
        r = urlparse(url, allow_fragments=False)
        bucket_name = r.netloc
        key = r.path.lstrip("/")
        if self.is_gce_instance():
            logger.debug("Generate signed URL for GCE instance")
            return self.python_cloud_function_get_signed_url(bucket_name, key)
        else:
            logger.debug("Generate signed URL for local instance")
            return self.generate_download_signed_url_v4(bucket_name, key)

    def generate_download_signed_url_v4(self, bucket_name, blob_name):
        """Generates a v4 signed URL for downloading a blob.

        Note that this method requires a service account key file. You can not use
        this if you are using Application Default Credentials from Google Compute
        Engine or from the Google Cloud SDK.
        """
        # bucket_name = 'your-bucket-name'
        # blob_name = 'your-object-name'

        client = self.get_client()
        bucket = self.get_bucket(client, bucket_name)
        blob = bucket.blob(blob_name)

        url = blob.generate_signed_url(
            version="v4",
            # This URL is valid for 15 minutes
            expiration=timedelta(minutes=self.presign_ttl),
            # Allow GET requests using this URL.
            method="GET",
        )

        logger.debug("Generated GCS signed url: " + url)
        return url

    def python_cloud_function_get_signed_url(self, bucket_name, blob_name):
        # https://gist.github.com/jezhumble/91051485db4462add82045ef9ac2a0ec
        # Copyright 2019 Google LLC.
        # SPDX-License-Identifier: Apache-2.0
        # This snippet shows you how to use Blob.generate_signed_url() from within compute engine / cloud functions
        # as described here: https://cloud.google.com/functions/docs/writing/http#uploading_files_via_cloud_storage
        # (without needing access to a private key)
        # Note: as described in that page, you need to run your function with a service account
        # with the permission roles/iam.serviceAccountTokenCreator
        auth_request = requests.Request()
        credentials, project = google.auth.default()
        storage_client = google_storage.Client(project, credentials)
        data_bucket = storage_client.lookup_bucket(bucket_name)
        signed_blob_path = data_bucket.blob(blob_name)
        expires_at_ms = datetime.now() + timedelta(minutes=self.presign_ttl)
        # This next line is the trick!
        signing_credentials = compute_engine.IDTokenCredentials(
            auth_request, "", service_account_email=None
        )
        signed_url = signed_blob_path.generate_signed_url(
            expires_at_ms, credentials=signing_credentials, version="v4"
        )
        return signed_url

    def resolve_uri(self, data):
        uri, storage = get_uri_via_regex(data, prefixes=(url_scheme,))
        if not storage:
            return
        logger.debug("Found matching storage uri in task data value: {uri}".format(uri=uri))
        resolved_uri = self.resolve_gs(uri)
        return data.replace(uri, resolved_uri)

    def scan_and_create_links(self):
        return self._scan_and_create_links(GCSImportStorageLink)


class GCSExportStorage(ExportStorage, GCSStorageMixin):
    def save_annotation(self, annotation):
        bucket = self.get_bucket()
        logger.debug(
            f"Creating new object on {self.__class__.__name__} Storage {self} for annotation {annotation}"
        )
        ser_annotation = self._get_serialized_data(annotation)
        with transaction.atomic():
            # Create export storage link
            link = GCSExportStorageLink.create(annotation, self)
            key = str(self.prefix) + "/" + link.key if self.prefix else link.key
            try:
                blob = bucket.blob(key)
                blob.upload_from_string(json.dumps(ser_annotation))
            except Exception as exc:
                logger.error(
                    f"Can't export annotation {annotation} to GCS storage {self}. Reason: {exc}",
                    exc_info=True,
                )


@receiver(post_save, sender=Annotation)
def export_annotation_to_gcs_storages(sender, instance, **kwargs):
    project = instance.task.project
    if hasattr(project, "io_storages_gcsexportstorages"):
        for storage in project.io_storages_gcsexportstorages.all():
            logger.debug(f"Export {instance} to GCS storage {storage}")
            storage.save_annotation(instance)


class GCSImportStorageLink(ImportStorageLink):
    storage = models.ForeignKey(GCSImportStorage, on_delete=models.CASCADE, related_name="links")


class GCSExportStorageLink(ExportStorageLink):
    storage = models.ForeignKey(GCSExportStorage, on_delete=models.CASCADE, related_name="links")



# Generated by Django 3.1.4 on 2021-05-17 21:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0010_auto_20210505_2037"),
    ]

    operations = [
        migrations.AlterField(
            model_name="project",
            name="sampling",
            field=models.CharField(
                choices=[
                    ("Sequential sampling", "Tasks are ordered by Data manager ordering"),
                    ("Uniform sampling", "Tasks are chosen randomly"),
                    (
                        "Uncertainty sampling",
                        "Tasks are chosen according to model uncertainty scores (active learning mode)",
                    ),
                ],
                default="Sequential sampling",
                max_length=100,
                null=True,
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import pytest
import json
import requests_mock
import math

from django.apps import apps
from django.urls import reverse
from tasks.models import Task, Annotation
from projects.models import Project
from .utils import invite_client_to_project, _client_is_annotator


@pytest.fixture
def configured_project_min_annotations_1(configured_project):
    p = Project.objects.get(id=configured_project.id)
    p.min_annotations_to_start_training = 1
    # p.agreement_method = p.SINGLE
    p.save()
    return p


@pytest.mark.django_db
@pytest.mark.parametrize(
    "result, logtext, ml_upload_called",
    [
        (
            json.dumps(
                [
                    {
                        "from_name": "text_class",
                        "to_name": "text",
                        "type": "labels",
                        "value": {"labels": ["class_A"], "start": 0, "end": 1},
                    }
                ]
            ),
            None,
            True,
        ),
        (json.dumps([]), None, True),
    ],
)
def test_create_annotation(
    caplog, any_client, configured_project_min_annotations_1, result, logtext, ml_upload_called
):
    task = Task.objects.first()
    if _client_is_annotator(any_client):
        assert invite_client_to_project(any_client, task.project).status_code == 200
    with requests_mock.Mocker() as m:
        m.post("http://localhost:8999/train")
        annotation = {"result": result, "task": task.id, "lead_time": 2.54}
        r = any_client.post(f"/api/tasks/{task.id}/annotations/", data=annotation)
        # check that submitted VALID data for task_annotation
        # makes task labeled
        task.refresh_from_db()
        assert task.is_labeled is True
        assert r.status_code == 201
        annotation = Annotation.objects.all()
        assert annotation.count() == 1
        annotation = annotation.first()
        assert annotation.task.id == task.id
        # annotator client
        if hasattr(any_client, "annotator") and any_client.annotator is not None:
            assert annotation.completed_by.id == any_client.user.id
        # business client
        else:
            assert annotation.completed_by.id == any_client.business.admin.id

        if apps.is_installed("businesses"):
            assert annotation.task.accuracy == 1.0

        if logtext:
            assert logtext in caplog.text


@pytest.mark.django_db
def test_create_annotation_with_ground_truth(
    caplog, any_client, configured_project_min_annotations_1
):

    task = Task.objects.first()
    if _client_is_annotator(any_client):
        assert invite_client_to_project(any_client, task.project).status_code == 200

    ground_truth = {
        "task": task.id,
        "result": json.dumps(
            [
                {
                    "from_name": "text_class",
                    "to_name": "text",
                    "value": {"labels": ["class_A"], "start": 0, "end": 1},
                }
            ]
        ),
        "ground_truth": True,
    }

    annotation = {
        "task": task.id,
        "result": json.dumps(
            [
                {
                    "from_name": "text_class",
                    "to_name": "text",
                    "value": {"labels": ["class_B"], "start": 0, "end": 1},
                }
            ]
        ),
    }

    with requests_mock.Mocker() as m:
        m.post("http://localhost:8999/train")

        # ground_truth doesn't affect statistics & ML backend
        r = any_client.post("/api/tasks/{}/annotations/".format(task.id), data=ground_truth)
        assert r.status_code == 201
        assert not m.called

        # real annotation triggers uploading to ML backend and recalculating accuracy
        r = any_client.post("/api/tasks/{}/annotations/".format(task.id), data=annotation)
        assert r.status_code == 201
        assert m.called
        task = Task.objects.get(id=task.id)
        assert task.annotations.count() == 2


@pytest.mark.django_db
def test_delete_annotation(business_client, configured_project):
    task = Task.objects.first()
    annotation = Annotation.objects.create(task=task, result=[])
    assert task.annotations.count() == 1
    r = business_client.delete("/api/annotations/{}/".format(annotation.id))
    assert r.status_code == 204
    assert task.annotations.count() == 0


@pytest.fixture
def annotations():
    task = Task.objects.first()
    return {
        "class_A": {
            "task": task.id,
            "result": json.dumps(
                [
                    {
                        "from_name": "text_class",
                        "to_name": "text",
                        "type": "labels",
                        "value": {"labels": ["class_A"], "start": 0, "end": 10},
                    }
                ]
            ),
        },
        "class_B": {
            "task": task.id,
            "result": json.dumps(
                [
                    {
                        "from_name": "text_class",
                        "to_name": "text",
                        "type": "labels",
                        "value": {"labels": ["class_B"], "start": 0, "end": 10},
                    }
                ]
            ),
        },
        "empty": {"task": task.id, "result": json.dumps([])},
    }


@pytest.fixture
def project_with_max_annotations_2(configured_project):
    configured_project.maximum_annotations = 2
    # configured_project.agreement_method = Project.SINGLE
    configured_project.save()


@pytest.mark.parametrize(
    "annotations_sequence, accuracy, is_labeled",
    [
        ([], None, False),
        ([("class_A", "business")], 1, False),
        ([("class_A", "annotator")], 1, False),
        ([("class_A", "business"), ("class_A", "business")], 1, True),
        ([("class_A", "business"), ("class_A", "annotator")], 1, True),
        ([("class_A", "annotator"), ("class_A", "business")], 1, True),
        ([("class_A", "business"), ("class_B", "business")], 0.5, True),
        ([("class_A", "business"), ("class_B", "annotator")], 0.5, True),
        ([("class_A", "annotator"), ("class_B", "business")], 0.5, True),
        ([("empty", "annotator"), ("empty", "business")], 1, True),
        ([("class_A", "annotator"), ("empty", "business")], 0.5, True),
    ],
)
@pytest.mark.django_db
def test_accuracy(
    business_client,
    annotator_client,
    project_with_max_annotations_2,
    annotations,
    annotations_sequence,
    accuracy,
    is_labeled,
):
    client = {"business": business_client, "annotator": annotator_client}
    task_id = next(iter(annotations.values()))["task"]
    task = Task.objects.get(id=task_id)
    invite_client_to_project(annotator_client, task.project)

    for annotation_key, client_key in annotations_sequence:
        r = client[client_key].post(
            reverse("tasks:api:task-annotations", kwargs={"pk": task_id}),
            data=annotations[annotation_key],
        )
        assert r.status_code == 201
    task = Task.objects.get(id=task_id)
    assert task.is_labeled == is_labeled


@pytest.mark.django_db
def test_accuracy_on_delete(business_client, project_with_max_annotations_2, annotations):
    task_id = next(iter(annotations.values()))["task"]
    for annotation in annotations.values():
        business_client.post(
            reverse("tasks:api:task-annotations", kwargs={"pk": task_id}), data=annotation
        )

    task = Task.objects.get(id=task_id)
    assert task.annotations.count() == len(annotations)
    assert task.is_labeled
    annotation_ids = [c.id for c in task.annotations.all()]
    r = business_client.delete(
        reverse("tasks:api-annotations:annotation-detail", kwargs={"pk": annotation_ids[0]})
    )
    assert r.status_code == 204
    task = Task.objects.get(id=task_id)
    assert task.is_labeled  # project.max_annotations = 2

    r = business_client.delete(
        reverse("tasks:api-annotations:annotation-detail", kwargs={"pk": annotation_ids[1]})
    )
    assert r.status_code == 204
    task = Task.objects.get(id=task_id)
    assert not task.is_labeled

    r = business_client.delete(
        reverse("tasks:api-annotations:annotation-detail", kwargs={"pk": annotation_ids[2]})
    )
    assert r.status_code == 204
    task = Task.objects.get(id=task_id)
    assert not task.is_labeled


# @pytest.mark.django_db
# def test_accuracy_on_delete(business_client, project_with_max_annotations_2, annotations):
#     task_id = next(iter(annotations.values()))['task']
#     for annotation in annotations.values():
#         business_client.post(reverse('tasks:api:task-annotations', kwargs={'pk': task_id}), data=annotation)
#
#     task = Task.objects.get(id=task_id)
#     assert task.annotations.count() == len(annotations)
#     if apps.is_installed('businesses'):
#         assert math.fabs(task.accuracy - 1 / 3) < 0.00001
#     assert task.is_labeled
#     annotation_ids = [c.id for c in task.annotations.all()]
#     r = business_client.delete(reverse('tasks:api-annotations:annotation-detail', kwargs={'pk': annotation_ids[0]}))
#     assert r.status_code == 204
#     task = Task.objects.get(id=task_id)
#     if apps.is_installed('businesses'):
#         assert task.accuracy == 0.5
#     assert task.is_labeled  # project.max_annotations = 2
#
#     r = business_client.delete(reverse('tasks:api-annotations:annotation-detail', kwargs={'pk': annotation_ids[1]}))
#     assert r.status_code == 204
#     task = Task.objects.get(id=task_id)
#     if apps.is_installed('businesses'):
#         assert task.accuracy == 1.0
#     assert not task.is_labeled
#
#     r = business_client.delete(reverse('tasks:api-annotations:annotation-detail', kwargs={'pk': annotation_ids[2]}))
#     assert r.status_code == 204
#     task = Task.objects.get(id=task_id)
#     if apps.is_installed('businesses'):
#         assert task.accuracy is None
#     assert not task.is_labeled



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.conf import settings as django_settings
from core.utils.common import collect_versions


def sentry_fe(request):
    # return the value you want as a dictionary, you may add multiple values in there
    return {"SENTRY_FE": django_settings.SENTRY_FE}


def settings(request):
    """Make available django settings on each template page"""
    versions = collect_versions()

    # django templates can't access names with hyphens
    versions["lsf"] = versions.get("label-studio-frontend", {})
    versions["lsf"]["commit"] = versions["lsf"].get("commit", "none")[0:6]

    if "backend" in versions:
        versions["backend"]["commit"] = versions["backend"].get("commit", "none")[0:6]

    return {"settings": django_settings, "versions": versions}



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import json
import logging

from django.db.models import Q, Avg, Count, Sum, Value, BooleanField, Case, When
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.db.models import JSONField
from django.core.validators import MinLengthValidator, MaxLengthValidator
from django.db import transaction, models
from annoying.fields import AutoOneToOneField

from rest_framework.exceptions import ValidationError

from tasks.models import (
    Task,
    Prediction,
    Annotation,
    Q_task_finished_annotations,
    Q_finished_annotations,
)
from core.utils.common import create_hash, pretty_date, sample_query, get_attr_or_item, load_func
from core.label_config import (
    parse_config,
    validate_label_config,
    extract_data_types,
    get_all_object_tag_names,
    config_line_stipped,
    get_sample_task,
    get_all_labels,
    get_all_control_tag_tuples,
    get_annotation_tuple,
)

logger = logging.getLogger(__name__)


class ProjectManager(models.Manager):
    def for_user(self, user):
        return self.filter(organization=user.active_organization)

    def with_counts(self):
        return self.annotate(
            task_number=Count("tasks", distinct=True),
            total_predictions_number=Count("tasks__predictions", distinct=True),
            total_annotations_number=Count(
                "tasks__annotations__id",
                distinct=True,
                filter=Q(tasks__annotations__was_cancelled=False),
            ),
            useful_annotation_number=Count(
                "tasks__annotations__id",
                distinct=True,
                filter=Q(tasks__annotations__was_cancelled=False)
                & Q(tasks__annotations__ground_truth=False)
                & Q(tasks__annotations__result__isnull=False),
            ),
            ground_truth_number=Count(
                "tasks__annotations__id",
                distinct=True,
                filter=Q(tasks__annotations__ground_truth=True),
            ),
            skipped_annotations_number=Count(
                "tasks__annotations__id",
                distinct=True,
                filter=Q(tasks__annotations__was_cancelled=True),
            ),
        )


ProjectMixin = load_func(settings.PROJECT_MIXIN)


class Project(ProjectMixin, models.Model):
    """ """

    objects = ProjectManager()
    __original_label_config = None

    title = models.CharField(
        _("title"),
        null=True,
        blank=True,
        default="",
        max_length=settings.PROJECT_TITLE_MAX_LEN,
        help_text=f"Project name. Must be between {settings.PROJECT_TITLE_MIN_LEN} and {settings.PROJECT_TITLE_MAX_LEN} characters long.",
        validators=[
            MinLengthValidator(settings.PROJECT_TITLE_MIN_LEN),
            MaxLengthValidator(settings.PROJECT_TITLE_MAX_LEN),
        ],
    )
    description = models.TextField(
        _("description"), blank=True, null=True, default="", help_text="Project description"
    )

    organization = models.ForeignKey(
        "organizations.Organization", on_delete=models.CASCADE, related_name="projects", null=True
    )
    label_config = models.TextField(
        _("label config"),
        blank=True,
        null=True,
        default="<View></View>",
        help_text="Label config in XML format. See more about it in documentation",
    )
    expert_instruction = models.TextField(
        _("expert instruction"),
        blank=True,
        null=True,
        default="",
        help_text="Labeling instructions in HTML format",
    )
    show_instruction = models.BooleanField(
        _("show instruction"),
        default=False,
        help_text="Show instructions to the annotator before they start",
    )

    show_skip_button = models.BooleanField(
        _("show skip button"),
        default=True,
        help_text="Show a skip button in interface and allow annotators to skip the task",
    )
    enable_empty_annotation = models.BooleanField(
        _("enable empty annotation"),
        default=True,
        help_text="Allow annotators to submit empty annotations",
    )

    show_annotation_history = models.BooleanField(
        _("show annotation history"),
        default=False,
        help_text="Show annotation history to annotator",
    )
    show_collab_predictions = models.BooleanField(
        _("show predictions to annotator"),
        default=True,
        help_text="If set, the annotator can view model predictions",
    )
    evaluate_predictions_automatically = models.BooleanField(
        _("evaluate predictions automatically"),
        default=False,
        help_text="Retrieve and display predictions when loading a task",
    )
    token = models.CharField(_("token"), max_length=256, default=create_hash, null=True, blank=True)
    result_count = models.IntegerField(
        _("result count"), default=0, help_text="Total results inside of annotations counter"
    )
    color = models.CharField(_("color"), max_length=16, default="#FFFFFF", null=True, blank=True)

    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="created_projects",
        on_delete=models.SET_NULL,
        null=True,
        verbose_name=_("created by"),
    )
    maximum_annotations = models.IntegerField(
        _("maximum annotation number"),
        default=1,
        help_text="Maximum number of annotations for one task. "
        "If the number of annotations per task is equal or greater "
        "to this value, the task is completed (is_labeled=True)",
    )
    min_annotations_to_start_training = models.IntegerField(
        _("min_annotations_to_start_training"),
        default=10,
        help_text="Minimum number of completed tasks after which model training is started",
    )

    control_weights = JSONField(
        _("control weights"), null=True, default=dict, help_text="Weights for control tags"
    )
    model_version = models.TextField(
        _("model version"),
        blank=True,
        null=True,
        default="",
        help_text="Machine learning model version",
    )
    data_types = JSONField(_("data_types"), default=dict, null=True)

    is_draft = models.BooleanField(
        _("is draft"),
        default=False,
        help_text="Whether or not the project is in the middle of being created",
    )
    is_published = models.BooleanField(
        _("published"),
        default=False,
        help_text="Whether or not the project is published to annotators",
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)

    SEQUENCE = "Sequential sampling"
    UNIFORM = "Uniform sampling"
    UNCERTAINTY = "Uncertainty sampling"

    SAMPLING_CHOICES = (
        (SEQUENCE, "Tasks are ordered by Data manager ordering"),
        (UNIFORM, "Tasks are chosen randomly"),
        (
            UNCERTAINTY,
            "Tasks are chosen according to model uncertainty scores (active learning mode)",
        ),
    )

    sampling = models.CharField(
        max_length=100, choices=SAMPLING_CHOICES, null=True, default=SEQUENCE
    )
    show_ground_truth_first = models.BooleanField(_("show ground truth first"), default=True)
    show_overlap_first = models.BooleanField(_("show overlap first"), default=True)
    overlap_cohort_percentage = models.IntegerField(_("overlap_cohort_percentage"), default=100)

    task_data_login = models.CharField(
        _("task_data_login"),
        max_length=256,
        blank=True,
        null=True,
        help_text="Task data credentials: login",
    )
    task_data_password = models.CharField(
        _("task_data_password"),
        max_length=256,
        blank=True,
        null=True,
        help_text="Task data credentials: password",
    )

    def __init__(self, *args, **kwargs):
        super(Project, self).__init__(*args, **kwargs)
        self.__original_label_config = self.label_config
        self.__maximum_annotations = self.maximum_annotations
        self.__overlap_cohort_percentage = self.overlap_cohort_percentage

        # TODO: once bugfix with incorrect data types in List
        # logging.warning('! Please, remove code below after patching of all projects (extract_data_types)')
        if self.label_config is not None:
            if self.data_types != extract_data_types(self.label_config):
                self.data_types = extract_data_types(self.label_config)

    @property
    def num_tasks(self):
        return self.tasks.count()

    def get_current_predictions(self):
        return Prediction.objects.filter(
            Q(task__project=self.id) & Q(model_version=self.model_version)
        )

    @property
    def num_predictions(self):
        return self.get_current_predictions().count()

    @property
    def num_annotations(self):
        return Annotation.objects.filter(
            Q(task__project=self) & Q_finished_annotations & Q(ground_truth=False)
        ).count()

    @property
    def has_predictions(self):
        return self.get_current_predictions().exists()

    @property
    def has_any_predictions(self):
        return Prediction.objects.filter(Q(task__project=self.id)).exists()

    @property
    def business(self):
        return self.created_by.business

    @property
    def is_private(self):
        return None

    @property
    def has_storages(self):
        return hasattr(self, "storages") and self.storages is not None and self.storages.count() > 0

    @property
    def secure_mode(self):
        return False

    @property
    def one_object_in_label_config(self):
        return len(self.data_types) <= 1

    @property
    def only_undefined_field(self):
        return (
            self.one_object_in_label_config
            and self.summary.common_data_columns
            and self.summary.common_data_columns[0] == settings.DATA_UNDEFINED_NAME
        )

    @property
    def get_labeled_count(self):
        return self.tasks.filter(is_labeled=True).count()

    @property
    def get_collected_count(self):
        return self.tasks.count()

    @property
    def num_tasks_with_annotations(self):
        return (
            self.tasks.filter(
                Q(annotations__isnull=False)
                & Q(annotations__ground_truth=False)
                & Q_task_finished_annotations
            )
            .distinct()
            .count()
        )

    @property
    def get_total_possible_count(self):
        """
            Tasks has overlap - how many tc should be accepted
            possible count = sum [ t.overlap for t in tasks]

        :return: N int total amount of Annotations that should be submitted
        """
        if self.tasks.count() == 0:
            return 0
        return self.tasks.aggregate(Sum("overlap"))["overlap__sum"]

    @property
    def get_available_for_labeling(self):
        return self.get_collected_count - self.get_labeled_count

    @property
    def need_annotators(self):
        return self.maximum_annotations - self.num_annotators

    @classmethod
    def find_by_invite_url(cls, url):
        token = url.strip("/").split("/")[-1]
        if len(token):
            return Project.objects.get(token=token)
        else:
            raise KeyError(f"Can't find Project by invite URL: {url}")

    def reset_token(self):
        self.token = create_hash()
        self.save()

    def add_collaborator(self, user):
        created = False
        with transaction.atomic():
            try:
                ProjectMember.objects.get(user=user, project=self)
            except ProjectMember.DoesNotExist:
                ProjectMember.objects.create(user=user, project=self)
                created = True
            else:
                logger.debug(f"Project membership {self} for user {user} already exists")
        return created

    def has_collaborator(self, user):
        return ProjectMember.objects.filter(user=user, project=self).exists()

    def has_collaborator_enabled(self, user):
        membership = ProjectMember.objects.filter(user=user, project=self)
        return membership.exists() and membership.first().enabled

    def update_tasks_states(
        self, maximum_annotations_changed, overlap_cohort_percentage_changed, tasks_number_changed
    ):

        # if only maximum annotations parameter is tweaked
        if maximum_annotations_changed and not overlap_cohort_percentage_changed:
            tasks_with_overlap = self.tasks.filter(overlap__gt=1)
            if tasks_with_overlap.exists():
                # if there is a part with overlaped tasks, affect only them
                tasks_with_overlap.update(overlap=self.maximum_annotations)
            else:
                # otherwise affect all tasks
                self.tasks.update(overlap=self.maximum_annotations)

        # if cohort slider is tweaked
        elif overlap_cohort_percentage_changed and self.maximum_annotations > 1:
            self._rearrange_overlap_cohort()

        # if adding/deleting tasks and cohort settings are applied
        elif (
            tasks_number_changed
            and self.overlap_cohort_percentage < 100
            and self.maximum_annotations > 1
        ):
            self._rearrange_overlap_cohort()

    def _rearrange_overlap_cohort(self):
        tasks_with_overlap = self.tasks.filter(overlap__gt=1)
        tasks_with_overlap_count = tasks_with_overlap.count()
        total_tasks = self.tasks.count()

        new_tasks_with_overlap_count = int(self.overlap_cohort_percentage / 100 * total_tasks + 0.5)
        if tasks_with_overlap_count > new_tasks_with_overlap_count:
            # TODO: warn if we try to reduce current cohort that is already labeled with overlap
            reduce_by = tasks_with_overlap_count - new_tasks_with_overlap_count
            reduce_tasks = sample_query(tasks_with_overlap, reduce_by)
            reduce_tasks.update(overlap=1)
            reduced_tasks_ids = reduce_tasks.values_list("id", flat=True)
            tasks_with_overlap.exclude(id__in=reduced_tasks_ids).update(
                overlap=self.maximum_annotations
            )

        elif tasks_with_overlap_count < new_tasks_with_overlap_count:
            increase_by = new_tasks_with_overlap_count - tasks_with_overlap_count
            tasks_without_overlap = self.tasks.filter(overlap=1)
            increase_tasks = sample_query(tasks_without_overlap, increase_by)
            increase_tasks.update(overlap=self.maximum_annotations)
            tasks_with_overlap.update(overlap=self.maximum_annotations)

    def remove_tasks_by_file_uploads(self, file_upload_ids):
        self.tasks.filter(file_upload_id__in=file_upload_ids).delete()

    def advance_onboarding(self):
        """Move project to next onboarding step"""
        po_qs = self.steps_left.order_by("step__order")
        count = po_qs.count()

        if count:
            po = po_qs.first()
            po.finished = True
            po.save()

            return count != 1

    def created_at_prettify(self):
        return self.created_at.strftime("%d %b %Y %H:%M:%S")

    def onboarding_step_finished(self, step):
        """Mark specific step as finished"""
        pos = ProjectOnboardingSteps.objects.get(code=step)
        po = ProjectOnboarding.objects.get(project=self, step=pos)
        po.finished = True
        po.save()

        return po

    def data_types_json(self):
        return json.dumps(self.data_types)

    def available_data_keys(self):
        return sorted(list(self.data_types.keys()))

    @classmethod
    def validate_label_config(cls, config_string):
        validate_label_config(config_string)

    def validate_config(self, config_string):
        self.validate_label_config(config_string)
        if not hasattr(self, "summary"):
            return

        # validate data columns consistency
        fields_from_config = get_all_object_tag_names(config_string)
        if not fields_from_config:
            logger.debug(f"Data fields not found in labeling config")
            return
        fields_from_data = set(self.summary.common_data_columns)
        fields_from_data.discard(settings.DATA_UNDEFINED_NAME)
        if fields_from_data and not fields_from_config.issubset(fields_from_data):
            different_fields = list(fields_from_config.difference(fields_from_data))
            raise ValidationError(
                f'These fields are not present in the data: {",".join(different_fields)}'
            )

        # validate annotations consistency
        annotations_from_config = set(get_all_control_tag_tuples(config_string))
        if not annotations_from_config:
            logger.debug(f"Annotation schema is not found in config")
            return
        annotations_from_data = set(self.summary.created_annotations)
        if annotations_from_data and not annotations_from_data.issubset(annotations_from_config):
            different_annotations = list(annotations_from_data.difference(annotations_from_config))
            diff_str = []
            for ann_tuple in different_annotations:
                from_name, to_name, t = ann_tuple.split("|")
                diff_str.append(
                    f"{self.summary.created_annotations[ann_tuple]} "
                    f"with from_name={from_name}, to_name={to_name}, type={t}"
                )
            diff_str = "\n".join(diff_str)
            raise ValidationError(
                f"Created annotations are incompatible with provided labeling schema, "
                f"we found:\n{diff_str}"
            )

        # validate labels consistency
        labels_from_config = get_all_labels(config_string)
        created_labels = self.summary.created_labels
        for control_tag_from_data, labels_from_data in created_labels.items():
            # Check if labels created in annotations, and their control tag has been removed
            if labels_from_data and control_tag_from_data not in labels_from_config:
                raise ValidationError(
                    f"There are {sum(labels_from_data.values(), 0)} annotation(s) created with tag "
                    f'"{control_tag_from_data}", you can\'t remove it'
                )
            labels_from_config_by_tag = set(labels_from_config[control_tag_from_data])
            if not set(labels_from_data).issubset(set(labels_from_config_by_tag)):
                different_labels = list(set(labels_from_data).difference(labels_from_config_by_tag))
                diff_str = "\n".join(
                    f"{l} ({labels_from_data[l]} annotations)" for l in different_labels
                )
                raise ValidationError(f"These labels still exist in annotations:\n{diff_str}")

    def _label_config_has_changed(self):
        return self.label_config != self.__original_label_config

    def delete_predictions(self):
        predictions = Prediction.objects.filter(task__project=self)
        count = predictions.count()
        predictions.delete()
        return {"deleted_predictions": count}

    def get_updated_weights(self):
        outputs = parse_config(self.label_config)
        control_weights = {}
        exclude_control_types = ("Filter",)
        for control_name in outputs:
            control_type = outputs[control_name]["type"]
            if control_type in exclude_control_types:
                continue
            control_weights[control_name] = {
                "overall": 1.0,
                "type": control_type,
                "labels": {label: 1.0 for label in outputs[control_name].get("labels", [])},
            }
        return control_weights

    def save(self, *args, recalc=True, **kwargs):
        exists = True if self.pk else False

        if self.label_config and (
            self._label_config_has_changed() or not exists or not self.control_weights
        ):
            self.control_weights = self.get_updated_weights()
        super(Project, self).save(*args, **kwargs)
        project_with_config_just_created = not exists and self.pk and self.label_config
        if self._label_config_has_changed() or project_with_config_just_created:
            self.data_types = extract_data_types(self.label_config)

        if self._label_config_has_changed():
            self.__original_label_config = self.label_config

        if not exists:
            steps = ProjectOnboardingSteps.objects.all()
            objs = [ProjectOnboarding(project=self, step=step) for step in steps]
            ProjectOnboarding.objects.bulk_create(objs)

        # argument for recalculate project task stats
        if recalc:
            self.update_tasks_states(
                maximum_annotations_changed=self.__maximum_annotations != self.maximum_annotations,
                overlap_cohort_percentage_changed=self.__overlap_cohort_percentage
                != self.overlap_cohort_percentage,
                tasks_number_changed=False,
            )
            self.__maximum_annotations = self.maximum_annotations
            self.__overlap_cohort_percentage = self.overlap_cohort_percentage

    def get_member_ids(self):
        if hasattr(self, "team_link"):
            # project has defined team scope
            # TODO: avoid checking team but rather add all project members when creating a project
            return self.team_link.team.members.values_list("user", flat=True)
        else:
            from users.models import User

            # TODO: may want to return all users from organization
            return User.objects.none()

    def has_team_user(self, user):
        return hasattr(self, "team_link") and self.team_link.team.has_user(user)

    def annotators(self):
        """Annotators connected to this project including team members"""
        from users.models import User

        member_ids = self.get_member_ids()
        team_members = User.objects.filter(id__in=member_ids).order_by("email")

        # add members from invited projects
        project_member_ids = self.members.values_list("user__id", flat=True)
        project_members = User.objects.filter(id__in=project_member_ids)

        annotators = team_members | project_members

        # set annotator.team_member=True if annotator is not an invited user
        annotators = annotators.annotate(
            team_member=Case(
                When(id__in=project_member_ids, then=Value(False)),
                default=Value(True),
                output_field=BooleanField(),
            )
        )
        return annotators

    def annotators_with_annotations(self, min_count=500):
        """Annotators with annotation number > min_number

        :param min_count: minimal annotation number to leave an annotators
        :return: filtered annotators
        """
        annotators = self.annotators()
        q = (
            Q(annotations__task__project=self)
            & Q_task_finished_annotations
            & Q(annotations__ground_truth=False)
        )
        annotators = annotators.annotate(
            annotation_count=Count("annotations", filter=q, distinct=True)
        )
        return annotators.filter(annotation_count__gte=min_count)

    def labeled_tasks(self):
        return self.tasks.filter(is_labeled=True)

    def has_annotations(self):
        from tasks.models import Annotation  # prevent cycling imports

        return Annotation.objects.filter(Q(task__project=self) & Q(ground_truth=False)).count() > 0

    # [TODO] this should be a template tag or something like this
    @property
    def label_config_line(self):
        c = self.label_config
        return config_line_stipped(c)

    def get_sample_task(self, label_config=None):
        config = label_config or self.label_config
        task, _, _ = get_sample_task(config)
        return task

    def pretty_model_version(self):
        if not self.model_version:
            return "Undefined model version"
        return pretty_date(self.model_version)

    def eta(self):
        """
            Show eta for project to be finished
            eta = avg task annotations finish time * remain annotations

            task has overlap = amount of task annotations to consider as finished (is_labeled)
            remain annotations = sum ( task annotations to be done to fulfill each unfinished task overlap)

        :return: time in seconds
        """
        # finished tasks * overlap
        finished_tasks = Task.objects.filter(project=self.id, is_labeled=True)
        # one could make more than need to overlap
        min_n_finished_annotations = sum([ft.overlap for ft in finished_tasks])

        annotations_unfinished_tasks = Annotation.objects.filter(
            task__project=self.id, task__is_labeled=False, ground_truth=False, result__isnull=False
        ).count()

        # get minimum remain annotations
        total_annotations_needed = self.get_total_possible_count
        annotations_remain = (
            total_annotations_needed - min_n_finished_annotations - annotations_unfinished_tasks
        )

        # get average time of all finished TC
        finished_annotations = Annotation.objects.filter(
            Q(task__project=self.id) & Q(ground_truth=False), result__isnull=False
        ).values("lead_time")
        avg_lead_time = finished_annotations.aggregate(avg_lead_time=Avg("lead_time"))[
            "avg_lead_time"
        ]

        if avg_lead_time is None:
            return None
        return avg_lead_time * annotations_remain

    def finished(self):
        return not self.tasks.filter(is_labeled=False).exists()

    def annotations_lead_time(self):
        annotations = Annotation.objects.filter(Q(task__project=self.id) & Q(ground_truth=False))
        return annotations.aggregate(avg_lead_time=Avg("lead_time"))["avg_lead_time"]

    @staticmethod
    def django_settings():
        return settings

    @staticmethod
    def max_tasks_file_size():
        return settings.TASKS_MAX_FILE_SIZE

    def get_control_tags_from_config(self):
        return parse_config(self.label_config)

    def get_parsed_config(self):
        return parse_config(self.label_config)

    def __str__(self):
        return f"{self.title} (id={self.id})" or _("Business number %d") % self.pk

    class Meta:
        db_table = "project"


class ProjectOnboardingSteps(models.Model):
    """ """

    DATA_UPLOAD = "DU"
    CONF_SETTINGS = "CF"
    PUBLISH = "PB"
    INVITE_EXPERTS = "IE"

    STEPS_CHOICES = (
        (DATA_UPLOAD, "Import your data"),
        (CONF_SETTINGS, "Configure settings"),
        (PUBLISH, "Publish project"),
        (INVITE_EXPERTS, "Invite collaborators"),
    )

    code = models.CharField(max_length=2, choices=STEPS_CHOICES, null=True)

    title = models.CharField(_("title"), max_length=1000, null=False)
    description = models.TextField(_("description"), null=False)
    order = models.IntegerField(default=0)

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)

    class Meta:
        ordering = ["order"]


class ProjectOnboarding(models.Model):
    """ """

    step = models.ForeignKey(
        ProjectOnboardingSteps, on_delete=models.CASCADE, related_name="po_through"
    )
    project = models.ForeignKey(Project, on_delete=models.CASCADE)

    finished = models.BooleanField(default=False)

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)

    def save(self, *args, **kwargs):
        super(ProjectOnboarding, self).save(*args, **kwargs)
        if ProjectOnboarding.objects.filter(project=self.project, finished=True).count() == 4:
            self.project.skip_onboarding = True
            self.project.save(recalc=False)


class ProjectMember(models.Model):

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="project_memberships",
        help_text="User ID",
    )  # noqa
    project = models.ForeignKey(
        Project, on_delete=models.CASCADE, related_name="members", help_text="Project ID"
    )
    enabled = models.BooleanField(default=True, help_text="Project member is enabled")
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)


class ProjectSummary(models.Model):

    project = AutoOneToOneField(
        Project, primary_key=True, on_delete=models.CASCADE, related_name="summary"
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, help_text="Creation time")

    # { col1: task_count_with_col1, col2: task_count_with_col2 }
    all_data_columns = JSONField(
        _("all data columns"),
        null=True,
        default=dict,
        help_text="All data columns found in imported tasks",
    )
    # [col1, col2]
    common_data_columns = JSONField(
        _("common data columns"),
        null=True,
        default=list,
        help_text="Common data columns found across imported tasks",
    )
    # { (from_name, to_name, type): annotation_count }
    created_annotations = JSONField(
        _("created annotations"),
        null=True,
        default=dict,
        help_text="Unique annotation types identified by tuple (from_name, to_name, type)",
    )  # noqa
    # { from_name: {label1: task_count_with_label1, label2: task_count_with_label2} }
    created_labels = JSONField(
        _("created labels"), null=True, default=dict, help_text="Unique labels"
    )

    def has_permission(self, user):
        return self.project.has_permission(user)

    def update_data_columns(self, tasks):
        common_data_columns = set()
        all_data_columns = dict(self.all_data_columns)
        for task in tasks:
            try:
                task_data = get_attr_or_item(task, "data")
            except KeyError:
                task_data = task
            task_data_keys = task_data.keys()
            for column in task_data_keys:
                all_data_columns[column] = all_data_columns.get(column, 0) + 1
            if not common_data_columns:
                common_data_columns = set(task_data_keys)
            else:
                common_data_columns &= set(task_data_keys)

        self.all_data_columns = all_data_columns
        if not self.common_data_columns:
            self.common_data_columns = list(sorted(common_data_columns))
        else:
            self.common_data_columns = list(
                sorted(set(self.common_data_columns) & common_data_columns)
            )
        logger.debug(f"summary.all_data_columns = {self.all_data_columns}")
        logger.debug(f"summary.common_data_columns = {self.common_data_columns}")
        self.save()

    def remove_data_columns(self, tasks):
        all_data_columns = dict(self.all_data_columns)
        keys_to_remove = []

        for task in tasks:
            task_data = get_attr_or_item(task, "data")
            for key in task_data.keys():
                if key in all_data_columns:
                    all_data_columns[key] -= 1
                    if all_data_columns[key] == 0:
                        keys_to_remove.append(key)
                        all_data_columns.pop(key)
        self.all_data_columns = all_data_columns

        if keys_to_remove:
            common_data_columns = list(self.common_data_columns)
            for key in keys_to_remove:
                if key in common_data_columns:
                    common_data_columns.remove(key)
            self.common_data_columns = common_data_columns
        logger.debug(f"summary.all_data_columns = {self.all_data_columns}")
        logger.debug(f"summary.common_data_columns = {self.common_data_columns}")
        self.save()

    def _get_annotation_key(self, result):
        result_type = result.get("type")
        if result_type in ("relation", "rating", "pairwise"):
            return None
        if "from_name" not in result or "to_name" not in result:
            logger.error(
                'Unexpected annotation.result format: "from_name" or "to_name" not found in %r'
                % result
            )
            return None
        result_from_name = result["from_name"]
        key = get_annotation_tuple(result_from_name, result["to_name"], result_type or "")
        return key

    def _get_labels(self, result):
        result_type = result.get("type")
        labels = []
        for label in result["value"].get(result_type, []):
            if isinstance(label, list):
                labels.extend(label)
            else:
                labels.append(label)
        return [str(l) for l in labels]

    def update_created_annotations_and_labels(self, annotations):
        created_annotations = dict(self.created_annotations)
        labels = dict(self.created_labels)
        for annotation in annotations:
            results = get_attr_or_item(annotation, "result") or []
            for result in results:

                # aggregate annotation types
                key = self._get_annotation_key(result)
                if not key:
                    continue
                created_annotations[key] = created_annotations.get(key, 0) + 1
                from_name = result["from_name"]

                # aggregate labels
                if from_name not in self.created_labels:
                    labels[from_name] = dict()

                for label in self._get_labels(result):
                    labels[from_name][label] = labels[from_name].get(label, 0) + 1

        logger.debug(f"summary.created_annotations = {created_annotations}")
        logger.debug(f"summary.created_labels = {labels}")
        self.created_annotations = created_annotations
        self.created_labels = labels
        self.save()

    def remove_created_annotations_and_labels(self, annotations):
        created_annotations = dict(self.created_annotations)
        labels = dict(self.created_labels)
        for annotation in annotations:
            results = get_attr_or_item(annotation, "result") or []
            for result in results:

                # reduce annotation counters
                key = self._get_annotation_key(result)
                if key in created_annotations:
                    created_annotations[key] -= 1
                    if created_annotations[key] == 0:
                        created_annotations.pop(key)

                # reduce labels counters
                from_name = result.get("from_name")
                if from_name not in labels:
                    continue
                for label in self._get_labels(result):
                    label = str(label)
                    if label in labels[from_name]:
                        labels[from_name][label] -= 1
                        if labels[from_name][label] == 0:
                            labels[from_name].pop(label)
                if not labels[from_name]:
                    labels.pop(from_name)
        logger.debug(f"summary.created_annotations = {created_annotations}")
        logger.debug(f"summary.created_labels = {labels}")
        self.created_annotations = created_annotations
        self.created_labels = labels
        self.save()



# Generated by Django 3.1.4 on 2021-03-11 05:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("io_storages", "0001_squashed_0002_auto_20210302_1827"),
    ]

    operations = [
        migrations.AddField(
            model_name="azureblobexportstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="azureblobexportstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
        migrations.AddField(
            model_name="azureblobimportstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="azureblobimportstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
        migrations.AddField(
            model_name="gcsexportstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="gcsexportstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
        migrations.AddField(
            model_name="gcsimportstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="gcsimportstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
        migrations.AddField(
            model_name="redisexportstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="redisexportstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
        migrations.AddField(
            model_name="redisimportstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="redisimportstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
        migrations.AddField(
            model_name="s3exportstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="s3exportstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
        migrations.AddField(
            model_name="s3importstorage",
            name="last_sync",
            field=models.DateTimeField(
                blank=True, help_text="Last sync finished time", null=True, verbose_name="last sync"
            ),
        ),
        migrations.AddField(
            model_name="s3importstorage",
            name="last_sync_count",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Count of tasks synced last time",
                null=True,
                verbose_name="last sync count",
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.urls import path, include

from . import api
from . import views


app_name = "data_export"


_api_urlpatterns = [
    # export api
    path("<int:pk>/export", api.ExportAPI.as_view(), name="project-export"),
    path(
        "<int:pk>/export/formats", api.ExportFormatsListAPI.as_view(), name="project-export-formats"
    ),
    # Previously exported results
    path("<int:pk>/export/files", api.ProjectExportFiles.as_view(), name="project-export-files"),
]

urlpatterns = [
    path("api/projects/", include((_api_urlpatterns, app_name), namespace="api-projects")),
    path(
        "api/auth/export/",
        api.ProjectExportFilesAuthCheck.as_view(),
        name="project-export-files-auth-check",
    ),
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import json
import logging

from django.shortcuts import redirect
from django.urls import reverse
from django.http import HttpResponseNotAllowed


logger = logging.getLogger(__name__)


class APIViewVirtualRedirectMixin(object):
    def post(self, request, *args, **kwargs):
        red = self.request.POST.get("_redirect", "").lower()
        method = self.request.POST.get("_method", "").lower()

        res = super(APIViewVirtualRedirectMixin, self).post(request, *args, **kwargs)

        if red == "true":
            if method == "delete" and hasattr(self, "redirect_delete_route"):
                return redirect(reverse(self.redirect_delete_route))
            elif hasattr(self, "redirect_route_callback"):
                return redirect(self.redirect_route_callback(**self.request.POST))
            elif hasattr(self, "redirect_kwarg"):
                return redirect(
                    reverse(self.redirect_route, kwargs={"pk": self.kwargs[self.redirect_kwarg]})
                )
            else:
                return redirect(reverse(self.redirect_route))
        else:
            return res


class APIViewVirtualMethodMixin(object):
    def post(self, request, *args, **kwargs):
        method = self.request.POST.get("_method", "").lower()

        if method == "put":
            res = self.put(request, *args, **kwargs)
        elif method == "delete":
            res = self.delete(request, *args, **kwargs)
        elif method == "patch":
            res = self.patch(request, *args, **kwargs)
        else:
            if hasattr(super(APIViewVirtualMethodMixin, self), "post"):
                res = super(APIViewVirtualMethodMixin, self).post(request, *args, **kwargs)
            else:
                return HttpResponseNotAllowed(permitted_methods=[])

        return res


class RequestDebugLogMixin(object):
    def initial(self, request, *args, **kwargs):
        try:
            logger.debug(f"Request to {request.path}:\n{json.dumps(request.data, indent=2)}")
        except Exception as exc:
            logger.error(exc, exc_info=True)
        super(RequestDebugLogMixin, self).initial(request, *args, **kwargs)


class DummyModelMixin:
    def has_permission(self, user):
        return True



import os
import sys
import colorama
import logging
import urllib.request


logger = logging.getLogger("main")


def start_fix():
    import zipfile
    import platform

    print(f"Copying sqlite3.dll to the current directory: {os.getcwd()} ... ", end="")

    work_dir = os.path.dirname(os.path.abspath(__file__))
    filename = (
        "sqlite-dll-win64-x64-3350500.zip"
        if platform.architecture()[0] == "64bit"
        else "sqlite-dll-win32-x86-3350500.zip"
    )

    url = "https://www.sqlite.org/2021/" + filename

    src = os.path.join(work_dir, "sqlite.zip")
    try:
        with urllib.request.urlopen(url) as f:
            with open(src, "wb") as f_out:
                f_out.write(f.read())
    except:
        print(
            colorama.Fore.LIGHTRED_EX
            + "\nCan't download sqlite.zip. Please, download it manually:\n"
            + url
        )
        print(colorama.Fore.WHITE)
        exit()

    with zipfile.ZipFile(src, "r") as zip_ref:
        zip_ref.extractall(".")

    print("finished")
    print(
        colorama.Fore.LIGHTGREEN_EX
        + "\nPlease restart Label Studio to load the updated sqlite.dll\n"
    )
    print(colorama.Fore.WHITE)
    exit()


def windows_dll_fix():
    """Copy sqlite.dll to the current directory and use it"""

    # check if it is not on windows
    if sys.platform != "win32":
        return
    print(f"Current platform is {sys.platform}, apply sqlite fix")

    # set env
    import ctypes

    path_to_dll = os.path.abspath(".")
    os.environ["PATH"] = path_to_dll + os.pathsep + os.environ["PATH"]
    try:
        ctypes.CDLL(os.path.join(path_to_dll, "sqlite3.dll"))
        print("Add current directory to PATH for DLL search: " + path_to_dll)
    except OSError:
        print("Can't load sqlite3.dll from current directory")

    # check sqlite version
    import sqlite3

    v = sqlite3.sqlite_version_info
    # if v[0] >= 3 and v[1] >= 35:
    #     print("sqlite3 version doesn't a fix")
    #     return

    # check python version and warn
    print("python version: {sys.version_info.major} sqlite minor version: {sys.version_info.minor}")
    if sys.version_info.major == 3 and sys.version_info.minor in [6, 7, 8]:
        print(
            "\n"
            + colorama.Fore.LIGHTYELLOW_EX
            + "You are on "
            + colorama.Fore.LIGHTRED_EX
            + f"Windows Python {sys.version_info.major}.{sys.version_info.minor}.\n"
            + colorama.Fore.LIGHTYELLOW_EX
            + f"This Python version uses SQLite "
            f"{colorama.Fore.LIGHTRED_EX}{v[0]}.{v[1]}.{v[2]} "
            + colorama.Fore.LIGHTYELLOW_EX
            + f"which does not support JSON Field.\n"
            + "Read more about this issue: "
            + colorama.Fore.LIGHTWHITE_EX
            + "https://code.djangoproject.com/wiki/JSON1Extension [Windows section]\n"
        )

        auto_agree = any([a == "--agree-fix-sqlite" for a in sys.argv])
        agree = "n"
        if not auto_agree:
            print(
                colorama.Fore.WHITE
                + "Label Studio can try to resolve this issue by downloading the correct "
                "sqlite.dll from https://sqlite.org in the current directory, "
                "do you want to proceed? \n [y/n] > ",
                end="",
            )
            agree = input()

        if agree == "y" or auto_agree:
            start_fix()

    print(colorama.Fore.WHITE)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from os.path import join
from django.conf import settings
from django.conf.urls import url, include
from django.urls import path, re_path
from django.views.static import serve
from rest_framework import routers

from users import views, api

router = routers.DefaultRouter()
router.register(r"users", api.UserAPI, basename="user")

urlpatterns = [
    url(r"^api/", include(router.urls)),
    # Authentication
    path("user/login/", views.user_login, name="user-login"),
    path("user/signup/", views.user_signup, name="user-signup"),
    path("user/account/", views.user_account, name="user-account"),
    url(r"^logout/?$", views.logout, name="logout"),
    # Password reset
    url(r"^password-reset/$", views.FPasswordResetView.as_view(), name="password_reset"),
    url(
        r"^password-reset/done/$",
        views.FPasswordResetDoneView.as_view(),
        name="password_reset_done",
    ),
    url(
        r"^password-reset/(?P<uidb64>[0-9A-Za-z_\-]+)/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$",
        views.FPasswordResetConfirmView.as_view(),
        name="password_reset_confirm",
    ),
    url(
        r"^password-reset/complete/$",
        views.FPasswordResetCompleteView.as_view(),
        name="password_reset_complete",
    ),
    # avatars
    re_path(
        r"^data/" + settings.AVATAR_PATH + "/(?P<path>.*)$",
        serve,
        kwargs={"document_root": join(settings.MEDIA_ROOT, settings.AVATAR_PATH)},
    ),
    # Token
    path(
        "api/current-user/reset-token/",
        api.UserResetTokenAPI.as_view(),
        name="current-user-reset-token",
    ),
    path("api/current-user/token", api.UserGetTokenAPI.as_view(), name="current-user-token"),
    path("api/current-user/whoami", api.UserWhoAmIAPI.as_view(), name="current-user-whoami"),
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import pytest
import json
import requests_mock

from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient
from .utils import ml_backend_mock

from projects.models import Project
from ml.models import MLBackendTrainJob
from tasks.models import Annotation, Task


@pytest.fixture
def client_and_token(business_client):
    token = Token.objects.get(user=business_client.business.admin)
    client = APIClient()
    client.credentials(HTTP_AUTHORIZATION="Token " + token.key)
    client.organization_pk = business_client.organization.pk
    return client, token


@pytest.fixture(params=["business_authorized", "user_with_token"])
def any_api_client(request, client_and_token, business_client):
    client, token = client_and_token
    result = {"type": request.param, "token": token}
    if request.param == "business_authorized":
        result["client"] = business_client
    elif request.param == "user_with_token":
        result["client"] = client
    return result


@pytest.mark.parametrize(
    "payload, response, status_code",
    [
        # status OK
        (
            {
                "title": "111",
                "label_config": '<View><Text name="my_text" value="$text"/><Choices name="my_class" toName="my_text"><Choice value="pos"/><Choice value="neg"/></Choices></View>',
            },  # noqa
            None,
            201,
        ),
        # invalid label config: unexisted toName
        (
            {
                "title": "111",
                "label_config": '<View><Text name="my_text" value="$text"/><Choices name="my_class" toName="unexisted"><Choice value="pos"/><Choice value="neg"/></Choices></View>',
            },  # noqa
            {"label_config": ["toName=\"unexisted\" not found in names: ['my_class', 'my_text']"]},
            400,
        ),
        # invalid label config: missed toName
        (
            {
                "title": "111",
                "label_config": '<View><Text name="my_text" value="$text"/><Choices name="my_class"><Choice value="pos"/><Choice value="neg"/></Choices></View>',
            },  # noqa
            {"label_config": ["Validation failed on : 'toName' is a required property"]},  # noqa
            400,
        ),
        # empty label config
        ({"title": "111", "label_config": None}, {"label_config": ["can only parse strings"]}, 400),
    ],
)
@pytest.mark.django_db
def test_create_project(client_and_token, payload, response, status_code):
    client, token = client_and_token
    payload["organization_pk"] = client.organization_pk
    with ml_backend_mock():
        r = client.post(
            "/api/projects/",
            data=json.dumps(payload),
            content_type="application/json",
            headers={"Authorization": f"Token {token}"},
        )
    assert r.status_code == status_code
    if response:
        response_data = r.json()
        if r.status_code == 400:
            assert response_data["validation_errors"] == response
        else:
            assert response_data == response


@pytest.mark.parametrize(
    "payload, response, status_code",
    [
        # status OK
        (
            {
                "label_config": '<View><Text name="my_text" value="$text"/><Choices name="my_class" toName="my_text"><Choice value="pos"/><Choice value="neg"/></Choices></View>'
            },  # noqa
            None,
            200,
        ),
        # TODO: this should instead of next one, but "configured_project" fixture doesn't update project.summary with data columns
        # invalid column
        # (
        #     {"label_config": "<View><Text name=\"my_text\" value=\"$text\"/><Choices name=\"my_class\" toName=\"my_text\"><Choice value=\"pos\"/><Choice value=\"neg\"/></Choices></View>"},
        #     # noqa
        #     {'label_config': ['These fields are not found in data: text']},
        #     400
        # ),
        # invalid label config: unexisted toName
        (
            {
                "label_config": '<View><Text name="my_text" value="$text"/><Choices name="my_class" toName="unexisted"><Choice value="pos"/><Choice value="neg"/></Choices></View>'
            },  # noqa
            {"label_config": ["toName=\"unexisted\" not found in names: ['my_class', 'my_text']"]},
            400,
        ),
        # invalid label config: missed toName
        (
            {
                "label_config": '<View><Text name="my_text" value="$text"/><Choices name="my_class"><Choice value="pos"/><Choice value="neg"/></Choices></View>'
            },  # noqa
            {"label_config": ["Validation failed on : 'toName' is a required property"]},  # noqa
            400,
        ),
    ],
)
@pytest.mark.django_db
def test_patch_project(client_and_token, configured_project, payload, response, status_code):
    client, token = client_and_token
    payload["organization_pk"] = client.organization_pk
    r = client.patch(
        f"/api/projects/{configured_project.id}/",
        data=json.dumps(payload),
        content_type="application/json",
        headers={"Authorization": f"Token {token}"},
    )
    assert r.status_code == status_code
    if response:
        response_data = r.json()
        if r.status_code == 400:
            assert response_data["validation_errors"] == response
        else:
            assert response_data == response


@pytest.mark.parametrize(
    "external_status_code, current_active_ml_backend_url, ml_backend_call_count",
    [
        (201, "http://my.super.ai", 4),
    ],
)
@pytest.mark.django_db
def test_creating_activating_new_ml_backend(
    client_and_token,
    configured_project,
    external_status_code,
    current_active_ml_backend_url,
    ml_backend_call_count,
):
    business_client, token = client_and_token
    with requests_mock.Mocker() as m:
        my_url = "http://my.super.ai"
        m.post(f"{my_url}/setup", text=json.dumps({"model_version": "Version from My Super AI"}))
        m.get(f"{my_url}/health", text=json.dumps({"status": "UP"}))
        r = business_client.post(
            "/api/ml",
            data=json.dumps(
                {"project": configured_project.id, "title": "My Super AI", "url": my_url}
            ),
            content_type="application/json",
            headers={"Authorization": f"Token {token}"},
        )
        assert r.status_code == external_status_code
        assert m.called
        assert m.call_count == ml_backend_call_count
        project = Project.objects.get(id=configured_project.id)
        all_urls = [m.url for m in project.ml_backends.all()]
        connected_ml = [url for url in all_urls if url == current_active_ml_backend_url]
        assert len(connected_ml) == 1, "\n".join(all_urls)


@pytest.mark.django_db
def test_delete_annotations(business_client, configured_project):
    business_client.delete(f"/api/projects/{configured_project.id}/annotations/")
    assert not Annotation.objects.filter(task__project=configured_project.id).exists()


# --- TaskAPI ---


@pytest.mark.parametrize(
    "response, status_code",
    [
        # status OK
        (
            {
                "annotations": [],
                "predictions": [],
                "drafts": [],
                "data": {"text": "text B", "meta_info": "meta info B"},
                "meta": {},
                "created_at": "",
                "updated_at": "",
                "is_labeled": False,
                "project": 0,
                "overlap": 1,
                "file_upload": None,
            },
            200,
        )
    ],
)
@pytest.mark.django_db
def test_get_task(client_and_token, configured_project, response, status_code):
    client, token = client_and_token
    task = configured_project.tasks.all()[0]
    response["project"] = configured_project.id
    response["created_at"] = task.created_at.isoformat().replace("+00:00", "Z")
    response["updated_at"] = task.updated_at.isoformat().replace("+00:00", "Z")
    response["id"] = task.id
    r = client.get(
        f"/api/tasks/{task.id}/",
        content_type="application/json",
        headers={"Authorization": f"Token {token}"},
    )
    assert r.status_code == status_code
    if response:
        assert r.json() == response


@pytest.mark.parametrize(
    "payload, response, status_code",
    [
        # status OK
        (
            {
                "annotations": [],
                "predictions": [],
                "data": {"text": "TEST1", "meta_info": "TEST2"},
                "meta": {},
                "created_at": "",
                "updated_at": "",
                "is_labeled": False,
                "project": 0,
                "file_upload": None,
            },
            {
                "id": 0,
                "annotations": [],
                "predictions": [],
                "data": {"text": "TEST1", "meta_info": "TEST2"},
                "meta": {},
                "created_at": "",
                "updated_at": "",
                "is_labeled": False,
                "project": 0,
                "overlap": 1,
                "file_upload": None,
            },
            200,
        )
    ],
)
@pytest.mark.django_db
def test_patch_task(client_and_token, configured_project, payload, response, status_code):
    client, token = client_and_token
    task = configured_project.tasks.all()[0]
    payload["project"] = configured_project.id

    r = client.patch(
        f"/api/tasks/{task.id}/",
        data=json.dumps(payload),
        content_type="application/json",
        headers={"Authorization": f"Token {token}"},
    )

    task = configured_project.tasks.all()[0]  # call DB again after update
    response["project"] = configured_project.id
    response["created_at"] = task.created_at.isoformat().replace("+00:00", "Z")
    response["updated_at"] = task.updated_at.isoformat().replace("+00:00", "Z")
    response["id"] = task.id

    assert r.status_code == status_code
    if response:
        assert r.json() == response
